#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A collection of classes and functions to support visualization of
data
"""
# Classes:
#     ChartConfig
#     ChartData

# Functions:
#     parse_instrument_data (Union[str, Path, pd.DataFrame]) -> pd.DataFrame
#     select_subset (list[str], pd.DataFrame) -> pd.DataFrame
#     plot_xy_chart (ChartData, ChartConfig) -> go.Figure:
#     create_single_plot (pd.DataFrame, pd.DataFrame, ChartConfig) -> go.Figure
#     create_subplots (pd.DataFrame, pd.DataFrame, ChartConfig) -> go.Figure:
#     create_overlay (pd.DataFrame, pd.DataFrame, ChartConfig) -> go.Figure
#     apply_single_config (go.Figure, ChartConfig)
#     apply_subplots_x_config (go.Figure, ChartConfig)
#     apply_subplots_y_config (go.Figure, ChartConfig)
#     apply_overlay_config (go.Figure, ChartConfig)
#     plot_ts_chart (
#         np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray,
#         np.ndarray, ChartConfig
#     ) -> go.Figure

# Native imports
import json
from logging import getLogger
from typing import Dict, List, Literal, Optional, Union
from pathlib import Path
import warnings

# Third-party imports
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly import subplots

# Internal imports
from seabirdscientific.interpret_sbs_variable import interpret_sbs_variable

# TODO: check python version
# TODO: translate warning and error messages

logger = getLogger(__name__)


class ChartConfig:
    """Dataclass to contain chart information and plotly settings"""

    def __init__(
        self,
        title: str,
        x_names: List[str],
        y_names: List[str],
        z_names: List[str],
        chart_type: Literal["overlay", "subplots"],
        bounds: Optional[Dict[Literal["x", "y", "z"], Dict[int, List[int]]]] = None,
        x_titles: Optional[List[str]] = None,
        y_titles: Optional[List[str]] = None,
        z_titles: Optional[List[str]] = None,
        plot_loop_edit_flags=False,
        lift_pen_over_bad_data=False,
        flag_value=-9.99e-29,
    ):
        """Initializes a chart config object to store the names, units,
        and data format.

        Config parameters must be in the same order as the data.

        :param title: Title of the chart
        :param x_names: X axis names
        :param y_names: Y axis names
        :param z_names: Z axis names
        :param chart_type: string to select type of chart "overlay" for
            multiple datasets sharing one axis, or "subplots" for
            multiple datasets on separate subplots.
        :param bounds: Chart axis bounds, for example:
            {'x': {0: [2, 10], 1: [1, 100]}
        :param x_titles: List of titles corresponding to x axis dataset.
            Defaults to units in line charts and empty string in ts plot
        :param y_titles: List of titles corresponding to y axis dataset.
            Defaults to units in line charts and empty string in ts plot
        :param z_titles: List of titles corresponding to z axis dataset.
            Defaults to units in line charts and empty string in ts plot
        :param plot_loop_edit_flags: If true, data are plotted
            regardless of loop edit flag values, otherwise data are not
            plotted where loop edit flags are equal to the flag_value
        :param lift_pen_over_bad_data: If true, flagged data are not
            drawn and instead leave gaps in plotted lines. Otherwise,
            lines are drawn between points surrounding flagged data (see
            SeaSoft manual, page 122)
        :param flag_value: A user configurable value that identifies
            flagged data. Defaults to -9.99e-29
        """

        self.title = title
        x_info = [interpret_sbs_variable(name) for name in x_names]
        y_info = [interpret_sbs_variable(name) for name in y_names]
        z_info = [interpret_sbs_variable(name) for name in z_names]
        self.x_names = [info["name"] for info in x_info]
        self.y_names = [info["name"] for info in y_info]
        self.z_names = [info["name"] for info in z_info]
        self.x_units = [info["units"] for info in x_info]
        self.y_units = [info["units"] for info in y_info]
        self.z_units = [info["units"] for info in z_info]
        self.chart_type = chart_type
        self.bounds = bounds if bounds is not None else {}
        self.x_titles = x_titles if x_titles is not None else x_names
        self.y_titles = y_titles if y_titles is not None else y_names
        self.z_titles = z_titles if z_titles is not None else z_names
        self.plot_loop_edit_flags = plot_loop_edit_flags
        self.lift_pen_over_bad_data = lift_pen_over_bad_data
        self.flag_value = flag_value

        axes: List[Literal["x", "y", "z"]] = ["x", "y", "z"]
        for axis in axes:
            if axis not in self.bounds.keys():
                self.bounds[axis] = {}


class ChartData:
    """Class to contain chart data and helper functions"""

    # TODO: move helper functions into ChartData

    def __init__(self, data_source: Union[str, pd.DataFrame], config: ChartConfig):
        """Initializes an object to store chart data.

        :param data_source: A file path (.csv, .asc, .json), a JSON
            string, or pandas DataFrame
        :param config: A ChartConfig object to configure plotly
        """

        data = parse_instrument_data(data_source)
        if data is not None:
            data.mask(data == config.flag_value, inplace=True)
            if not config.plot_loop_edit_flags and "flag" in data.columns:
                mask = data["flag"].isnull()
                data.loc[mask, :] = np.nan
            self.x = select_subset(config.x_names, data)
            self.y = select_subset(config.y_names, data)
            self.z = select_subset(config.z_names, data)


def parse_instrument_data(source: Union[str, Path, pd.DataFrame]) -> pd.DataFrame:
    """Top level function for converting instrument data to numpy array.

    Currently supports pandas dataframes, json strings, or a Path to the
    following file types: .csv, .asc (comma separated only), .json.

    :param source: A JSON string, file path (.csv, .asc, .json), or
        pandas DataFrame

    :return: pandas dataframe containing field names and data

    """

    try:
        if isinstance(source, pd.DataFrame):
            data = source.copy()

        elif isinstance(source, Path):
            suffix = source.suffix.lower()
            if suffix in (".csv", ".asc"):
                data = pd.read_csv(source)
            elif suffix == ".json":
                with open(source, mode="r") as js_data:
                    data = pd.DataFrame.from_dict(json.load(js_data), orient="columns")

        elif isinstance(source, str):
            data = pd.DataFrame.from_dict(json.loads(source), orient="columns")

        else:
            raise TypeError

        if "data" not in locals():
            raise NameError

        if not isinstance(data, pd.DataFrame):
            raise TypeError

        columns = [interpret_sbs_variable(column)["name"] for column in data.columns]
        for old_column, new_column in zip(data.columns, columns):
            data.rename(columns={old_column: new_column}, inplace=True)
        return data

    except (NameError, TypeError) as e:
        logger.error(e)
        return None


def select_subset(axis_names: list[str], data: pd.DataFrame) -> pd.DataFrame:
    """Takes a list of axis names and returns a data set for each name
    in the list.

    If axis_names is empty the function will return a DataFrame of
    integers representing the sample count of the data. This could be
    used in a single series chart for example.

    Otherwise, the function will return a DataFrame for each name in the
    list. This would be for a single xy chart or an overlay/subplot
    chart.

    Example:

    data = read_data("./example.csv")

    subset = select_subset(["T090C", "C0Sm"], data)

    :param axis_names: List of axis names corresponding to the data
    :param data: The numpy DataFrame returned from read_data()

    :return: A tuple with the axis name and data

    """

    if len(axis_names) == 0:
        return pd.DataFrame({"Sample Count": list(range(0, len(data)))})

    return data[axis_names]


def plot_xy_chart(data: ChartData, config: ChartConfig) -> go.Figure:
    """Takes instrument data and a config and plots an XY chart with one
    or more data sets.

    :param data: Data object with x, y, z, data selected
        according to the config
    :param config: Config object with various plotly
        settings

    :return: A plotly.graph_objects.Figure
    """

    figure = go.Figure()

    # single data set
    if len(data.x.columns) == 1 and len(data.y.columns) == 1:
        figure = create_single_plot(data.x, data.y, config)

    # too many data sets
    elif len(data.x.columns) > 1 and len(data.y.columns) > 1:
        message = "Only one axis can support multiple data sets"
        logger.warning(message)
        warnings.warn(message)

    # multiple data sets
    elif len(data.x.columns) > 1 or len(data.y.columns) > 1:
        if config.chart_type == "overlay":
            figure = create_overlay(data.x, data.y, config)
        elif config.chart_type == "subplots":
            figure = create_subplots(data.x, data.y, config)

    else:
        # getting here should not be possible unless data and config are
        # altered outside their init functions
        raise ValueError

    if not config.lift_pen_over_bad_data:
        figure.update_traces(connectgaps=True)

    return figure


def create_single_plot(x: pd.DataFrame, y: pd.DataFrame, config: ChartConfig) -> go.Figure:
    """Creates a single XY plot, with one or more data sets.

    If there are multiple datasets for the x or y axis, an overlay plot
    will be generated

    :param x: Numpy array of data for the x axis
    :param y: Numpy array of data for the y axis
    :param config: Dataclass with settings for the plotly chart

    :return: A plotly.graph_objects.Figure displaying the provided x y
        data
    """

    if any(name in x.columns for name in y.columns):
        logger.warning("Duplicate data names will be omitted")

    figure = px.line(
        data_frame=pd.concat([x, y], axis=1),
        x=x.columns[0] if len(x.columns) == 1 else x.columns,
        y=y.columns[0] if len(y.columns) == 1 else y.columns,
        title=config.title,
        # markers=True,
    )

    apply_single_config(figure, config)

    return figure


def create_subplots(x: pd.DataFrame, y: pd.DataFrame, config: ChartConfig) -> go.Figure:
    """Creates a chart with multiple subplots.

    :param x: Pandas DataFrame of data for the x axis
    :param y: Pandas DataFrame of data for the y axis
    :param config: Dataclass with settings for the plotly chart

    :return: A plotly.graph_objects.Figure with multiple subplots
    """

    if any(name in x.columns for name in y.columns):
        logger.warning("Duplicate data names will be omitted")

    figure = go.Figure()
    figure.layout.title = config.title

    if len(x.columns) > 1 and len(y.columns) == 1:
        figure = subplots.make_subplots(
            rows=len(y.columns),
            cols=len(x.columns),
            column_titles=list(x.columns),
            y_title=y.columns[0],
            figure=figure,
        )
        column = 1
        for x_column in x.columns:
            figure.add_trace(
                go.Scatter(
                    x=x[x_column],
                    y=y[y.columns[0]],
                    name=x_column,
                ),
                row=1,
                col=column,
            )
            column += 1
        apply_subplots_x_config(figure, config)

    elif len(x.columns) == 1 and len(y.columns) > 1:
        figure = subplots.make_subplots(
            rows=len(y.columns),
            cols=len(x.columns),
            row_titles=list(y.columns),
            x_title=x.columns[0],
            figure=figure,
        )
        row = 1
        for y_column in y.columns:
            figure.add_trace(
                go.Scatter(
                    x=x[x.columns[0]],
                    y=y[y_column],
                    name=y_column,
                ),
                row=row,
                col=1,
            )
            row += 1
        apply_subplots_y_config(figure, config)

    return figure


def create_overlay(x: pd.DataFrame, y: pd.DataFrame, config: ChartConfig) -> go.Figure:
    """Creates a chart with multiple datasets overlayed on one axis.

    :param x: Pandas DataFrame of data for the x axis
    :param y: Pandas DataFrame of data for the y axis
    :param config: Dataclass with settings for the plotly chart

    :return: A plotly.graph_objects.Figure
    """

    if any(name in x.columns for name in y.columns):
        logger.warning("Duplicate data names will be omitted")

    figure = go.Figure()
    figure.layout.title = config.title

    if len(x.columns) > 1 and len(y.columns) == 1:
        x_axis = 1
        for x_column in x.columns:
            figure.add_trace(
                go.Scatter(x=x[x_column], y=y[y.columns[0]], name=x_column, xaxis=f"x{x_axis}")
            )
            x_axis += 1

        apply_overlay_config(figure, config)

    elif len(x.columns) == 1 and len(y.columns) > 1:
        y_axis = 1
        for y_column in y.columns:
            figure.add_trace(
                go.Scatter(x=x[x.columns[0]], y=y[y_column], name=y_column, yaxis=f"y{y_axis}")
            )
            y_axis += 1

        apply_overlay_config(figure, config)

    return figure


def apply_single_config(figure: go.Figure, config: ChartConfig):
    """Updates various chart settings for single plots.

    :param figure: The figure being updated
    :param config: The user defined config being applied to the figure
    """

    figure.update_layout(
        xaxis={
            "title": "" if len(config.x_titles) < 1 else config.x_titles[0],
            "domain": [max(0, 0.1 * (len(config.y_titles) - 1)), 1],
            "range": None if len(config.bounds["x"]) < 1 else config.bounds["x"][0],
        },
        yaxis={
            "title": "" if len(config.y_titles) < 1 else config.y_titles[0],
            "domain": [max(0, 0.1 * (len(config.x_titles) - 1)), 1],
            "range": None if len(config.bounds["y"]) < 1 else config.bounds["y"][0],
        },
    )


def apply_subplots_x_config(figure: go.Figure, config: ChartConfig):
    """Updates various chart settings for charts with multiple x axes.

    Config parameters may contain upto 4 arguments per axis, and must be
    in the same order as the data. Hence all of the magic number
    indexing below

    :param figure: The figure being updated
    :param config: The user defined config being applied to the figure
    """

    y_range = None if len(config.bounds["y"]) < 1 else config.bounds["y"][0]

    figure.update_layout(
        xaxis={
            "title": "" if len(config.x_titles) < 1 else config.x_titles[0],
            "range": None if len(config.bounds["x"]) < 1 else config.bounds["x"][0],
        },
        xaxis2={
            "title": "" if len(config.x_titles) < 2 else config.x_titles[1],
            "range": None if len(config.bounds["x"]) < 2 else config.bounds["x"][1],
        },
        xaxis3={
            "title": "" if len(config.x_titles) < 3 else config.x_titles[2],
            "range": None if len(config.bounds["x"]) < 3 else config.bounds["x"][2],
        },
        xaxis4={
            "title": "" if len(config.x_titles) < 4 else config.x_titles[3],
            "range": None if len(config.bounds["x"]) < 4 else config.bounds["x"][3],
        },
        yaxis={
            "title": "" if len(config.y_titles) < 1 else config.y_titles[0],
            "range": y_range,
        },
        yaxis2={"range": y_range},
        yaxis3={"range": y_range},
        yaxis4={"range": y_range},
    )


def apply_subplots_y_config(figure: go.Figure, config: ChartConfig):
    """Updates various chart settings for charts with multiple y axes.

    Config parameters may contain upto 4 arguments per axis, and must be
    in the same order as the data. Hence all of the magic number
    indexing below

    :param figure: The figure being updated
    :param config: The user defined config being applied to the figure
    """

    x_range = None if len(config.bounds["x"]) < 1 else config.bounds["x"][0]

    figure.update_layout(
        xaxis={"range": x_range},
        xaxis2={"range": x_range},
        xaxis3={"range": x_range},
        xaxis4={
            "title": "" if len(config.x_titles) < 1 else config.x_titles[0],
            "range": x_range,
        },
        yaxis={
            "title": "" if len(config.y_titles) < 1 else config.y_titles[0],
            "range": None if len(config.bounds["y"]) < 1 else config.bounds["y"][0],
        },
        yaxis2={
            "title": "" if len(config.y_titles) < 2 else config.y_titles[1],
            "range": None if len(config.bounds["y"]) < 2 else config.bounds["y"][1],
        },
        yaxis3={
            "title": "" if len(config.y_titles) < 3 else config.y_titles[2],
            "range": None if len(config.bounds["y"]) < 3 else config.bounds["y"][2],
        },
        yaxis4={
            "title": "" if len(config.y_titles) < 4 else config.y_titles[3],
            "range": None if len(config.bounds["y"]) < 3 else config.bounds["y"][3],
        },
    )


def apply_overlay_config(figure: go.Figure, config: ChartConfig):
    """Updates various chart settings for charts with multiple y axes.

    Config parameters may contain upto 4 arguments per axis, and must be
    in the same order as the data.
    Hence all of the magic number indexing below

    :param figure: The figure being updated
    :param config: The user defined config being applied to the figure
    """

    figure.update_layout(
        xaxis={
            "title": "" if len(config.x_titles) < 1 else config.x_titles[0],
            "domain": [max(0, 0.1 * (len(config.y_titles) - 1)), 1],
            "range": None if len(config.bounds["x"]) < 1 else config.bounds["x"][0],
            "title_standoff": 5,
        },
        xaxis2={
            "title": "" if len(config.x_titles) < 2 else config.x_titles[1],
            "overlaying": "x",
            "position": 0.1,
            "range": None if len(config.bounds["x"]) < 2 else config.bounds["x"][1],
            "title_standoff": 5,
        },
        xaxis3={
            "title": "" if len(config.x_titles) < 3 else config.x_titles[2],
            "overlaying": "x",
            "position": 0.2,
            "range": None if len(config.bounds["x"]) < 3 else config.bounds["x"][2],
            "title_standoff": 5,
        },
        xaxis4={
            "title": "" if len(config.x_titles) < 4 else config.x_titles[3],
            "overlaying": "x",
            "position": 0.3,
            "range": None if len(config.bounds["x"]) < 4 else config.bounds["x"][3],
            "title_standoff": 5,
        },
        yaxis={
            "title": "" if len(config.y_titles) < 1 else config.y_titles[0],
            "domain": [0.1 * (len(config.x_titles)), 1],
            "position": 0,
            "range": None if len(config.bounds["y"]) < 1 else config.bounds["y"][0],
            "title_standoff": 5,
        },
        yaxis2={
            "title": "" if len(config.y_titles) < 2 else config.y_titles[1],
            "overlaying": "y",
            "position": 0.1,
            "range": None if len(config.bounds["y"]) < 2 else config.bounds["y"][1],
            "title_standoff": 5,
        },
        yaxis3={
            "title": "" if len(config.y_titles) < 3 else config.y_titles[2],
            "overlaying": "y",
            "position": 0.2,
            "range": None if len(config.bounds["y"]) < 3 else config.bounds["y"][2],
            "title_standoff": 5,
        },
        yaxis4={
            "title": "" if len(config.y_titles) < 4 else config.y_titles[3],
            "overlaying": "y",
            "position": 0.3,
            "range": None if len(config.bounds["y"]) < 3 else config.bounds["y"][3],
            "title_standoff": 5,
        },
    )


def plot_ts_chart(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    x_vec: np.ndarray,
    y_vec: np.ndarray,
    z_mat: np.ndarray,
    config: ChartConfig,
) -> go.Figure:
    """Overlays a scatter plot onto a contour plot to create a TS plot.
    Takes as args the xyz properties on a Contour object. In a future
    version these will be replaced with a single contour object

    :param x: absolute salinity
    :param y: conservative temperature
    :param z: potential density
    :param x_vec: absolute salinity vector
    :param y_vec: conservative temperature vector
    :param z_mat: potential density matrix
    :param config: Config object with key/values required by conversion
        function
    :return: A plotly.graph_objects.Figure
    """

    if len(config.x_names) > 1 or len(config.y_names) > 1 or len(config.z_names) > 1:
        logger.warning(
            "plot_ts_chart expects one data set for each axis. Extra data sets are ignored"
        )

    # Create 2 plots using plotly (not plotly express)
    # T-S diagram with x-y values colored by z
    trace1 = go.Scatter(
        x=x,
        y=y,
        mode="markers",
        marker={
            "color": z,
            "showscale": True,
            "size": 2,
            "colorbar": {
                # sigma_theta kg m-3
                "title": {
                    "text": config.z_titles[0],
                    "side": "top",
                }
            },
        },
    )

    # Contours in gray
    colorscale = [[0, "gray"], [1, "gray"]]
    trace2 = go.Contour(
        x=x_vec,
        y=y_vec,
        z=z_mat,
        showscale=False,
        colorscale=colorscale,
        contours={"coloring": "lines", "showlabels": True},
    )

    # Overlay the plots
    fig = subplots.make_subplots()
    fig.add_trace(trace1)
    fig.add_trace(trace2)

    # Centered title, axis labels
    fig.update_layout(
        height=600,
        width=600,
        title_text=config.title,
        title_x=0.5,
        xaxis_title=config.x_titles[0],
        yaxis_title=config.y_titles[0],
    )

    return fig



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""Calibration coefficient objects for data conversion"""

# Native imports
from dataclasses import dataclass
from typing import Optional
import warnings

# Third-party imports

# Sea-Bird imports

# Internal imports


@dataclass
class TemperatureCoefficients:
    """
    :param a0: coefficient
    :param a1: coefficient
    :param a2: coefficient
    :param a3: coefficient
    """

    a0: float
    a1: float
    a2: float
    a3: float


@dataclass
class TemperatureFrequencyCoefficients:
    """
    :param g: coefficient
    :param h: coefficient
    :param i: coefficient
    :param j: coefficient
    :param f0: coefficient
    """

    g: float
    h: float
    i: float
    j: float
    f0: float


@dataclass
class PressureCoefficients:
    """
    :param pa0: coefficient
    :param pa1: coefficient
    :param pa2: coefficient
    :param ptca0: coefficient
    :param ptca1: coefficient
    :param ptca2: coefficient
    :param ptcb0: coefficient
    :param ptcb1: coefficient
    :param ptcb2: coefficient
    :param ptempa0: coefficient
    :param ptempa1: coefficient
    :param ptempa2: coefficient
    """

    pa0: float
    pa1: float
    pa2: float
    ptca0: float
    ptca1: float
    ptca2: float
    ptcb0: float
    ptcb1: float
    ptcb2: float
    ptempa0: float
    ptempa1: float
    ptempa2: float


@dataclass
class PressureDigiquartzCoefficients:
    """
    :param a0: coefficient
    :param a1: coefficient
    :param a2: coefficient
    :param b0: coefficient
    :param b1: coefficient
    :param b2: coefficient
    :param c0: coefficient
    :param c1: coefficient
    :param c2: coefficient
    :param d0: coefficient
    :param d1: coefficient
    :param d2: coefficient
    """

    c1: float
    c2: float
    c3: float
    d1: float
    d2: float
    t1: float
    t2: float
    t3: float
    t4: float
    t5: float
    AD590M: float
    AD590B: float

    def __init__(
        self,
        c1: float,
        c2: float,
        c3: float,
        d1: float,
        d2: float,
        t1: float,
        t2: float,
        t3: float,
        t4: float,
        t5: float,
        AD590M: Optional[float] = None,
        AD590B: Optional[float] = None,
    ):
        self.c1 = c1
        self.c2 = c2
        self.c3 = c3
        self.d1 = d1
        self.d2 = d2
        self.t1 = t1
        self.t2 = t2
        self.t3 = t3
        self.t4 = t4
        self.t5 = t5
        if AD590M is not None:
            self.AD590M = AD590M
        else:
            self.AD590M = 0.028927  # default value for SBE16plus V2 and SBE19plus V2

        if AD590B is not None:
            self.AD590B = AD590B
        else:
            self.AD590B = -41.1733  # default value for SBE16plus V2 and SBE19plus V2


@dataclass
class ConductivityCoefficients:
    """
    :param g: coefficient
    :param h: coefficient
    :param i: coefficient
    :param j: coefficient
    :param cpcor: compressibility coefficient
    :param ctcor: coefficient
    :param wbotc: bridge oscillator temperature coefficient see the
        37 Manual: https://www.seabird.com/asset-get.download.jsa?id=54627862348
    """

    g: float
    h: float
    i: float
    j: float
    cpcor: float
    ctcor: float
    wbotc: float


@dataclass
class ECOCoefficients:
    """
    :param slope: units/count for digital, units/V for analog
    :param offset: dark counts for digital, dark voltage for analog
    """

    slope: float
    offset: float


@dataclass
class Oxygen63Coefficients:
    """
    :param a0: coefficient
    :param a1: coefficient
    :param a2: coefficient
    :param b0: coefficient
    :param b1: coefficient
    :param c0: coefficient
    :param c1: coefficient
    :param c2: coefficient
    :param e: coefficient
    """

    a0: float
    a1: float
    a2: float
    b0: float
    b1: float
    c0: float
    c1: float
    c2: float
    e: float


@dataclass
class Thermistor63Coefficients:
    """
    :param ta0: coefficient
    :param ta1: coefficient
    :param ta2: coefficient
    :param ta3: coefficient
    """

    ta0: float
    ta1: float
    ta2: float
    ta3: float


@dataclass
class Oxygen43Coefficients:
    """
    :param soc: linear scaling coefficient
    :param v_offset: voltage at zero oxygen signal
    :param tau_20: sensor time constant tau(T,P) at 20 C, 1 atmosphere, 0 PSU;
        slope term in calculation of tau(T,P)
    :param a: coefficient
    :param b: coefficient
    :param c: coefficient
    :param e: coefficient
    :param d0: tau(T,P) coefficient
    :param d1: tau(T,P) coefficient
    :param d2: tau(T,P) coefficient
    :param h1: hysteresis correction coefficient
    :param h2: hysteresis correction coefficient
    :param h3: hysteresis correction coefficient
    """

    soc: float
    v_offset: float
    tau_20: float
    a: float
    b: float
    c: float
    e: float
    d0: float  # TODO: the sensor lists a d0 coefficient, but it doesn't seem to be used?
    d1: float
    d2: float
    h1: float
    h2: float
    h3: float


@dataclass
class PH18Coefficients:
    """
    :param slope: calibration slope
    :param offset: calibration offset
    """

    slope: float
    offset: float


@dataclass
class PARCoefficients:
    """
    :param im: immersion coefficient
    :param a0: calibration slope
    :param a1: calibration offset
    :param multiplier: 1.0 for units of μmol photons/m2*s
    """

    im: float
    a0: float
    a1: float
    multiplier: float


@dataclass
class SPARCoefficients:
    """
    :param im: immersion coefficient
    :param a0: calibration slope
    :param a1: calibration offset
    :param conversion_factor: 1.0 for units of μmol photons/m2*s
    """

    im: float
    a0: float
    a1: float
    conversion_factor: float


@dataclass
class TemperatureSeaFETCoefficients:
    """
    :param tdfa0: temeperature coefficient
    :param tdfa1: temeperature coefficient
    :param tdfa2: temeperature coefficient
    :param tdfa3: temeperature coefficient
    """

    tdfa0: float
    tdfa1: float
    tdfa2: float
    tdfa3: float


class PHSeaFETInternalCoefficients:
    """
    :param kdf0: internal K0 coefficient
    :param kdf2: internal K2 coefficient

    :param k0: Deprecated, use kdf0
    :param k2: Deprecated, use kdf2
    :param int_k0: Deprecated, use kdf0
    :param int_k2: Deprecated, use kdf2
    """

    def __init__(
        self,
        kdf0: float = 0,
        kdf2: float = 0,
        k0: Optional[float] = None,
        k2: Optional[float] = None,
        int_k0: Optional[float] = None,
        int_k2: Optional[float] = None,
    ):
        self.kdf0 = kdf0
        self.kdf2 = kdf2
        if k0 is not None:
            self.k0 = k0
            self.kdf0 = k0
            warnings.warn("k0 is deprecated, use kdf0", DeprecationWarning)
        if k2 is not None:
            self.k2 = k2
            self.kdf2 = k2
            warnings.warn("k2 is deprecated, use kdf2", DeprecationWarning)
        if int_k0 is not None:
            self.int_k0 = int_k0
            self.kdf0 = int_k0
            warnings.warn("int_k0 is deprecated, use kdf0", DeprecationWarning)
        if int_k2 is not None:
            self.int_k2 = int_k2
            self.kdf2 = int_k2
            warnings.warn("int_k2 is deprecated, use kdf2", DeprecationWarning)


class PHSeaFETExternalCoefficients:
    """
    :param k0: external K0 coefficient
    :param k2: external K2 coefficient
    :param k2_poly_order: order of K2 pressure compensation calculation
    :param k2f0: f(P) coefficient
    :param k2f1: f(P) coefficient
    :param k2f2: f(P) coefficient
    :param k2f3: f(P) coefficient
    :param fp_poly_order: order of pressure compensation calculation
    :param f0: f(P) coefficient
    :param f1: f(P) coefficient
    :param f2: f(P) coefficient
    :param f3: f(P) coefficient
    :param f4: f(P) coefficient
    :param f5: f(P) coefficient
    :param f6: f(P) coefficient
    """

    def __init__(
        self,
        k0: float = 0,
        k2: float = 0,
        k2_poly_order=0,
        k2f0: float = 0,
        k2f1: float = 0,
        k2f2: float = 0,
        k2f3: float = 0,
        fp_poly_order: float = 0,
        f0: float = 0,
        f1: float = 0,
        f2: float = 0,
        f3: float = 0,
        f4: float = 0,
        f5: float = 0,
        f6: float = 0,
        ext_k0: Optional[float] = None,
        ext_k2: Optional[float] = None,
    ):
        self.k0 = k0
        self.k2 = k2
        self.k2_poly_order = k2_poly_order
        self.k2f0 = k2f0
        self.k2f1 = k2f1
        self.k2f2 = k2f2
        self.k2f3 = k2f3
        self.fp_poly_order = fp_poly_order
        self.f0 = f0
        self.f1 = f1
        self.f2 = f2
        self.f3 = f3
        self.f4 = f4
        self.f5 = f5
        self.f6 = f6

        if ext_k0 is not None:
            self.ext_k0 = ext_k0
            self.k0 = ext_k0
            warnings.warn("ext_k0 is deprecated, use k0", DeprecationWarning)
        if ext_k2 is not None:
            self.ext_k2 = ext_k2
            self.k2 = ext_k2
            warnings.warn("ext_k2 is deprecated, use k2", DeprecationWarning)


@dataclass
class AltimeterCoefficients:
    """
    :param slope: coefficient
    :param offset: coefficient
    """

    slope: float
    offset: float



"""instrument data conversion unit tests."""

# Native imports
from datetime import datetime
from pathlib import Path

# Third-party imports

# Sea-Bird imports

# Internal imports
import seabirdscientific.instrument_data as id

test_data = Path("./tests/resources/test-data")


class TestCnvToInstrumentData:
    measurement_series = id.MeasurementSeries(
        label="test",
        description="",
        units="",
        start_time=datetime(2000, 1, 1, 1, 1, 1),
        values=[10, 20, 30],
    )

    def test_create_measurement_series_pass(self):
        assert self.measurement_series.label == "test"

    def test_create_instrument_data_pass(self):
        data = id.InstrumentData(
            measurements={"test": self.measurement_series},
            interval_s=None,
            latitude=0.0,
            start_time=None,
            sample_count=None,
        )
        assert data is not None  # fails before assert if data fails to instantiate

    def test_cnv_to_instrument_data_pass(self):
        file_path = test_data / "SBE37SM.cnv"
        data = id.cnv_to_instrument_data(file_path)

        expected_labels = [
            "tv290C",
            "cond0S/m",
            "prdM",
            "prdE",
            "depSM",
            "depSF",
            "sal00",
            "flag",
        ]
        for expected, result in zip(expected_labels, data.measurements):
            assert expected == result

        assert data.interval_s == 120

        assert data.measurements["tv290C"].label == "tv290C"
        assert data.measurements["tv290C"].description == "Temperature"
        assert data.measurements["tv290C"].units == "ITS-90, deg C"
        assert data.measurements["tv290C"].start_time == datetime(2014, 2, 19, 13, 30, 1)
        assert data.measurements["tv290C"].values[0] == 20.8989
        assert data.measurements["tv290C"].values[-1] == -45.7713

    def test_duplicate_labels(self):
        """
        Verify that duplicate labels are handled correctly.
        This .cnv file is from the Australian Institute of Marine Science
        """
        file_path = test_data / "WQR086CFACLWDB.cnv"
        data = id.cnv_to_instrument_data(file_path)

        expected_labels = [
            "prDM",
            "t090C",
            "c0S/m",
            "scan",
            "depSM",
            "timeS",
            "cpar",
            "spar",
            "turbWETntu0",
            "flECO-AFL",
            "par",
            "t190C",
            "c1S/m",
            "depSM_1",  # this duplicate "depSM" was a problem, increments to "depSM1"
            "sal00",
            "nbin",
            "flag",
        ]

        assert list(data.measurements.keys()) == expected_labels

        # MeasurementSeries should have the original label
        assert data.measurements["depSM_1"].label == "depSM"

        first_values = [m.values[0] for m in data.measurements.values()]
        first_data_line = "1.006 28.0831 5.811991 5248 0.998 218.616 6.9285e+01 3.9379e+02 11.8057 0.8223 2.7316e+02 28.0829 5.811250 1.000 36.2666 31 0.0000e+00"
        expected_first_values = [float(v) for v in first_data_line.split()]
        assert first_values == expected_first_values


class TestReadHex19plus:
    filepath = test_data / "19plus_V2_trim.hex"
    raw = id.read_hex_file(
        filepath,
        id.InstrumentType.SBE19Plus,
        [
            id.Sensors.Temperature,
            id.Sensors.Conductivity,
            id.Sensors.Pressure,
            id.Sensors.ExtVolt0,
            id.Sensors.ExtVolt1,
            id.Sensors.ExtVolt2,
            id.Sensors.ExtVolt4,
        ],
    )

    def test_read_19plus_hex(self):
        assert self.raw["temperature"].iloc[0] == 322798
        assert self.raw["temperature"].iloc[-1] == 326317
        assert self.raw["conductivity"].iloc[0] == 675415 / 256
        assert self.raw["conductivity"].iloc[-1] == 1590943 / 256
        assert self.raw["pressure"].iloc[0] == 533539
        assert self.raw["pressure"].iloc[-1] == 533538
        assert self.raw["temperature compensation"].iloc[0] == 20625 / 13107
        assert self.raw["temperature compensation"].iloc[-1] == 20361 / 13107


class TestReadHex39plus:
    filepath = test_data / "SBE39plus.hex"
    raw = id.read_hex_file(
        filepath,
        id.InstrumentType.SBE39Plus,
        [
            id.Sensors.Temperature,
            id.Sensors.Conductivity,
            id.Sensors.Pressure,
        ],
    )

    def test_read_39plus_hex(self):
        assert self.raw["temperature"].iloc[0] == 1804057
        assert self.raw["temperature"].iloc[-1] == 1803910
        assert self.raw["pressure"].iloc[0] == 108184
        assert self.raw["pressure"].iloc[-1] == 108132
        assert self.raw["temperature compensation"].iloc[0] == 445677321 / 13107
        assert self.raw["temperature compensation"].iloc[-1] == 445664214 / 13107
        assert self.raw["date time"].iloc[0] == datetime(2025, 9, 15, 20, 27, 22)
        assert self.raw["date time"].iloc[-1] == datetime(2025, 9, 15, 20, 28, 9)


class TestReadSBE911plus:
    filepath = test_data / "911plus_trim.hex"
    raw = id.read_hex_file(
        filepath,
        id.InstrumentType.SBE911Plus,
        [
            id.Sensors.Temperature,
            id.Sensors.Conductivity,
            id.Sensors.Pressure,
            id.Sensors.SecondaryTemperature,
            id.Sensors.SecondaryConductivity,
            id.Sensors.ExtVolt0,
            id.Sensors.ExtVolt1,
            id.Sensors.ExtVolt2,
            id.Sensors.ExtVolt3,
            id.Sensors.ExtVolt4,
            id.Sensors.ExtVolt5,
            id.Sensors.ExtVolt6,
            id.Sensors.ExtVolt7,
            id.Sensors.SPAR,
            id.Sensors.nmeaLocation,
            id.Sensors.SystemTime,
        ],
    )

    # Expected values are from SBE Data Processing
    def test_read_911plus_hex(self):
        assert round(self.raw["temperature"].iloc[0], 3) == 4651.168
        assert round(self.raw["conductivity"].iloc[0], 3) == 6638.984
        assert round(self.raw["digiquartz pressure"].iloc[0], 3) == 33302.258
        assert round(self.raw["secondary temperature"].iloc[0], 3) == 4651.781
        assert round(self.raw["secondary conductivity"].iloc[0], 3) == 6230.480
        assert round(self.raw["volt 0"].iloc[0], 4) == 3.7338
        assert round(self.raw["volt 1"].iloc[0], 4) == 0.2112
        assert round(self.raw["volt 2"].iloc[0], 4) == 4.9328
        assert round(self.raw["volt 3"].iloc[0], 4) == 2.3370
        assert round(self.raw["volt 4"].iloc[0], 4) == 2.4884
        assert round(self.raw["volt 5"].iloc[0], 4) == 2.6777
        assert round(self.raw["volt 6"].iloc[0], 4) == 0.1502
        assert round(self.raw["volt 7"].iloc[0], 4) == 2.7912
        assert round(self.raw["surface par"].iloc[0], 5) == 1.43101
        assert self.raw["NMEA Latitude"].iloc[0] == 34.2757
        assert self.raw["NMEA Longitude"].iloc[0] == -120.0256
        assert (
            int(self.raw["SBE911 pump status"].iloc[0]) == 0
        )  # Does not match SBE Processing, matches Fathom
        assert (
            int(self.raw["SBE911 bottom contact status"].iloc[0]) == 0
        )  # Does not match SBE Processing, matches Fathom
        assert int(self.raw["SBE911 confirm status"].iloc[0]) == 1
        assert int(self.raw["SBE911 modem status"].iloc[0]) == 1
        assert int(self.raw["data integrity"].iloc[0]) == 83
        assert self.raw["system time"].iloc[0] == datetime(2025, 7, 28, 23, 43, 53)


class TestReadSeaFET2:
    filepath = test_data / "SeaFET2.hex"
    raw = id.read_hex_file(filepath, id.InstrumentType.SeaFET2, [], False, True)

    def test_read_seaFET2_hex(self):
        assert self.raw["vrs external"].iloc[0] == 5106181
        assert self.raw["vrs internal"].iloc[0] == 5096276
        assert self.raw["ph temperature"].iloc[0] == 1002679
        assert self.raw["vk"].iloc[0] == 5960288
        assert self.raw["ib"].iloc[0] == 8381912
        assert self.raw["ik"].iloc[0] == 8384541
        assert self.raw["relative humidity"].iloc[0] == 17632
        assert self.raw["internal temperature"].iloc[0] == 26016
        assert self.raw["error flag"].iloc[0] == 0
        assert self.raw["date time"].iloc[0] == datetime(2025, 9, 15, 21, 35, 48)


class TestReadNMEAHex:
    filepath = test_data / "19plus_V2_CTD_nmea_data.hex"
    raw = id.read_hex_file(
        filepath,
        id.InstrumentType.SBE19Plus,
        [
            id.Sensors.Temperature,
            id.Sensors.Conductivity,
            id.Sensors.Pressure,
            id.Sensors.ExtVolt0,
            id.Sensors.ExtVolt1,
            id.Sensors.nmeaLatitude,
            id.Sensors.nmeaLongitude,
            id.Sensors.statusAndSign,
            id.Sensors.nmeaTime,
        ],
    )

    def test_read_nmea_hex(self):
        assert self.raw["NMEA Latitude"].iloc[0] == 48.62280000
        assert self.raw["NMEA Latitude"].iloc[-1] == 48.62298000
        assert self.raw["NMEA Longitude"].iloc[0] == -123.50002000
        assert self.raw["NMEA Longitude"].iloc[-1] == -123.49994000
        assert self.raw["NMEA Date Time"].iloc[0] == datetime(2023, 3, 18, 12, 15, 19)
        assert self.raw["NMEA Date Time"].iloc[-1] == datetime(2023, 3, 18, 12, 36, 2)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Native imports
import re

# Third-party imports

# Sea-Bird imports

# Internal imports


def interpret_sbs_variable(sbs_var):
    """SBE variable name conversion to matlab friendly strings.

    Converts the strings that define SBE variables to strings that
    exclude special characters. Gives units for each variable and
    defines whether it is a number or a text string.

    All variables imported from SBEDataProcessing_7.26.4.pdf. There may
    be additional variables added in future versions of Sea-Bird Data
    processing modules that are not included here.

    :param sbs_var: Name of the dataset

    :return: The name with special characters removed, the data format,
        and the units. kvar_name: converted, matlab code readable
        variable name. kvar_format: defines whether variable is a float
        or a string. kvar_units: variable units. Empty string if
        undefined.
    """

    # TEMPERATURE
    # primary sensor
    kvar_format = r"%f"

    if sbs_var in {"t090Cm", "t4990C", "tnc90C", "tv290C", "t090C"}:
        kvar_name = "t090C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t090F", "t4990F", "tnc90F", "tv290F"}:
        kvar_name = "t090F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t068C", "t4968C", "tnc68C", "tv268C"}:
        kvar_name = "t068C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t068F", "t4968F", "tnc68F", "tv268F"}:
        kvar_name = "t068F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"t090"}:
        kvar_name = "t090"
        kvar_units = "SBE49, deg C"

    # secondary temperature
    elif sbs_var in {"t190C", "tnc290C"}:
        kvar_name = "t190C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t190F", "tnc290F"}:
        kvar_name = "t190F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t168C", "tnc268C"}:
        kvar_name = "t168C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t168F", "tnc268F"}:
        kvar_name = "t168F"
        kvar_units = "ITS-68, deg F"

    # SBE38 REFERENCE TEMPERATURE
    # primary sensor
    elif sbs_var in {"t3890C", "t38_90C"}:
        kvar_name = "t03890C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t3890F", "t38_90F"}:
        kvar_name = "t03890F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t3868C", "t38_68C"}:
        kvar_name = "t03868C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t3868F", "t38_68F"}:
        kvar_name = "t03868F"
        kvar_units = "ITS-68, deg F"
    # secondary sensor
    elif sbs_var in {"t3890C1"}:
        kvar_name = "t13890C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t3890F1"}:
        kvar_name = "t13890F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t3868C1"}:
        kvar_name = "t13868C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t3868F1"}:
        kvar_name = "t13868F"
        kvar_units = "ITS-68, deg F"

    # TEMPERATURE DIFFERENCE
    elif sbs_var in {"T2-T190C"}:
        kvar_name = "tdiff90C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"T2-T190F"}:
        kvar_name = "tdiff90F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"T2-T168C"}:
        kvar_name = "tdiff68C"
        kvar_units = "IPTS-68, deg C"
    elif sbs_var in {"T2-T168F"}:
        kvar_name = "tdiff68F"
        kvar_units = "IPTS-68, deg F"

    # CONDUCTIVITY
    # primary conductivity
    elif sbs_var in {"c_S/m", "cond0S/m", "c0S/m"}:
        kvar_name = "c0Sm"
        kvar_units = "S/m"
    elif sbs_var in {"c_mS/cm", "cond0mS/cm", "c0mS/cm"}:
        kvar_name = "c0mScm"
        kvar_units = "mS/cm"
    elif sbs_var in {"c_uS/cm", "cond0uS/cm", "c0uS/cm"}:
        kvar_name = "c0uScm"
        kvar_units = "microS/cm"
    # secondary conductivity
    elif sbs_var in {"c1S/m"}:
        kvar_name = "c1Sm"
        kvar_units = "S/m"
    elif sbs_var in {"c1mS/cm"}:
        kvar_name = "c1mScm"
        kvar_units = "mS/cm"
    elif sbs_var in {"c1uS/cm"}:
        kvar_name = "c1uScm"
        kvar_units = "microS/cm"

    # CONDUCTIVITY DIFFERENCE
    elif sbs_var in {"C2-C1S/m"}:
        kvar_name = "cdiffSm"
        kvar_units = "S/m"
    elif sbs_var in {"C2-C1mS/cm"}:
        kvar_name = "cdiffmScm"
        kvar_units = "mS/cm"
    elif sbs_var in {"C2-C1uS/cm"}:
        kvar_name = "cdiffuScm"
        kvar_units = "microS/cm"

    # PRESSURE
    # user entered pressure values
    elif sbs_var in {"pr"}:  # SBE 49
        kvar_name = "p"
        kvar_units = "db"
    elif sbs_var in {"prM"}:
        kvar_name = "pM"
        kvar_units = "db (user entered)"
    elif sbs_var in {"prE"}:
        kvar_name = "pE"
        kvar_units = "psi (user entered)"
    # digiquartz pressure
    elif sbs_var in {"prDM"}:
        kvar_name = "pm"
        kvar_units = "db"
    elif sbs_var in {"prDE"}:
        kvar_name = "pe"
        kvar_units = "psi"
    # strain gauge pressure
    elif sbs_var in {"prSM", "prdM"}:
        kvar_name = "pm"
        kvar_units = "db"
    elif sbs_var in {"prSE", "prdE"}:
        kvar_name = "pe"
        kvar_units = "psi"
    # pressure temperature
    elif sbs_var in {"ptempC"}:
        kvar_name = "ptempC"
        kvar_units = "deg C"
    elif sbs_var in {"ptempF"}:
        kvar_name = "ptempF"
        kvar_units = "deg F"
    # FGP Pressure
    elif sbs_var in {"fgp0"}:
        kvar_name = "fgp0"
        kvar_units = "KPa"
    elif sbs_var in {"fgp1"}:
        kvar_name = "fgp1"
        kvar_units = "KPa"
    elif sbs_var in {"fgp2"}:
        kvar_name = "fgp2"
        kvar_units = "KPa"
    elif sbs_var in {"fgp3"}:
        kvar_name = "fgp3"
        kvar_units = "KPa"
    elif sbs_var in {"fgp4"}:
        kvar_name = "fgp4"
        kvar_units = "KPa"
    elif sbs_var in {"fgp5"}:
        kvar_name = "fgp5"
        kvar_units = "KPa"
    elif sbs_var in {"fgp6"}:
        kvar_name = "fgp6"
        kvar_units = "KPa"
    elif sbs_var in {"fgp7"}:
        kvar_name = "fgp7"
        kvar_units = "KPa"
    # SBE 50
    elif sbs_var in {"pr50M"}:
        kvar_name = "p050m"
        kvar_units = "db"
    elif sbs_var in {"pr50E"}:
        kvar_name = "p050e"
        kvar_units = "psi"
    elif sbs_var in {"pr50M1"}:
        kvar_name = "p150m"
        kvar_units = "db"
    elif sbs_var in {"pr50E1"}:
        kvar_name = "p150e"
        kvar_units = "psi"

    # FREQUENCY OUTPUT
    elif sbs_var in {
        "f0",
        "f1",
        "f2",
        "f3",
        "f4",
        "f5",
        "f6",
        "f7",
        "f8",
        "f9",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
        "f19",
        "f20",
        "f21",
        "f22",
        "f23",
        "f24",
        "f25",
        "f26",
        "f27",
        "f28",
        "f29",
        "f30",
        "f31",
        "f32",
        "f33",
        "f34",
        "f35",
        "f36",
    }:
        kvar_name = sbs_var
        kvar_units = "Hz"

    # VOLTAGE OUTPUT
    elif sbs_var in {
        "v0",
        "v1",
        "v2",
        "v3",
        "v4",
        "v5",
        "v6",
        "v7",
        "v8",
        "v9",
        "v10",
        "v11",
        "v12",
        "v13",
        "v14",
        "v15",
    }:
        kvar_name = sbs_var
        kvar_units = "V"

    # GTD-DO SENSORS
    # PRESSURE
    elif sbs_var in {"GTDDOP0", "GTDDOP1", "GTDDOPdiff"}:
        kvar_name = sbs_var
        kvar_units = "mb"
    # TEMPERATURE
    elif sbs_var in {"GTDDOT0", "GTDDOT1", "GTDDOTdiff"}:
        kvar_name = sbs_var
        kvar_units = "deg C"
    # GTD-N2 SENSOR
    # PRESSURE
    elif sbs_var in {"GTDN2P0", "GTDN2P1", "GTDN2Pdiff"}:
        kvar_name = sbs_var
        kvar_units = "mb"
    # TEMPERATURE
    elif sbs_var in {"GTDN2T0", "GTDN2T1", "GTDN2Tdiff"}:
        kvar_name = sbs_var
        kvar_units = "deg C"

    # TIME
    # elapsed time
    elif sbs_var in {"timeS"}:
        kvar_name = "timeS"
        kvar_units = "elapsed time seconds"
    elif sbs_var in {"timeM"}:
        kvar_name = "timeM"
        kvar_units = "elapsed time minutes"
    elif sbs_var in {"timeH"}:
        kvar_name = "timeH"
        kvar_units = "elapsed time hours"
    elif sbs_var in {"timeJ"}:
        kvar_name = "timeJ"
        kvar_units = "elapsed time Julian Days"
    elif sbs_var in {"timeN"}:
        kvar_name = "timeN"
        kvar_units = "From NMEA: seconds since January 1, 2000"
    elif sbs_var in {"timeQ"}:
        kvar_name = "timeQ"
        kvar_units = "From NMEA: seconds since January 1, 1970"
    elif sbs_var in {"timeK"}:
        kvar_name = "timeK"
        kvar_units = "SBE timestamp: seconds since January 1, 2000"
    elif sbs_var in {"timeJV2"}:  # julian days
        kvar_name = "jdays"
        kvar_units = "SBE timestamp: julian days"
    elif sbs_var in {"timeSCP"}:  # julian days
        kvar_name = "jdays"
        kvar_units = "SBE timestamp: julian days"
    elif sbs_var in {"timeY"}:
        kvar_name = "timeY"
        kvar_units = "Computer Time: seconds since January 1, 1970"

    # POSITION
    elif sbs_var in {"latitude"}:
        kvar_name = "lat"
        kvar_units = "from NMEA degrees"
    elif sbs_var in {"longitude"}:
        kvar_name = "lon"
        kvar_units = "from NMEA degrees"

    # DEPTH
    elif sbs_var in {"depSM"}:
        kvar_name = "depSM"
        kvar_units = "depth salt water, m"
    elif sbs_var in {"depSF"}:
        kvar_name = "depSF"
        kvar_units = "depth salt water, ft"
    elif sbs_var in {"depFM"}:
        kvar_name = "depFM"
        kvar_units = "depth fresh water, m"
    elif sbs_var in {"depFF"}:
        kvar_name = "depSM"
        kvar_units = "depth fresh water, ft"
    elif sbs_var in {"dNMEA"}:
        kvar_name = "depSM"
        kvar_units = "depth from NMEA salt water, m"

    # POTENTIAL TEMPERATURE
    # I use the abbreviation of the greek symbol theta to designate
    # potential temperature
    elif sbs_var in {"potemp090C"}:
        kvar_name = "th090C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"potemp090F"}:
        kvar_name = "th090F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"potemp068C"}:
        kvar_name = "th068C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"potemp068F"}:
        kvar_name = "th068F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"potemp190C"}:
        kvar_name = "th190C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"potemp190F"}:
        kvar_name = "th190F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"potemp168C"}:
        kvar_name = "th168C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"potemp168F"}:
        kvar_name = "th168F"
        kvar_units = "ITS-68, deg F"

    # POTENTIAL TEMPEATURE DIFFERENCE
    elif sbs_var in {"potemp90Cdiff"}:
        kvar_name = "th090Cdiff"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"potemp90Fdiff"}:
        kvar_name = "th090Fdiff"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"potemp68Cdiff"}:
        kvar_name = "th068Cdiff"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"potemp68Fdiff"}:
        kvar_name = "th068Fdiff"
        kvar_units = "ITS-68, deg F"

    # POTENTIAL TEMPEATURE ANOMALY
    elif sbs_var in {"tha090C"}:
        kvar_name = "tha090C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"tha090F"}:
        kvar_name = "tha090F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"tha068C"}:
        kvar_name = "tha068C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"tha068F"}:
        kvar_name = "tha068F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"tha190C"}:
        kvar_name = "tha190C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"tha190F"}:
        kvar_name = "tha190F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"tha168C"}:
        kvar_name = "tha168C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"tha168F"}:
        kvar_name = "tha168F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"pta090C"}:
        kvar_name = "pta090C"
        kvar_units = "ITS-90, deg C"

    # SALINITY
    elif sbs_var in {"sal00"}:
        kvar_name = "psal0"
        kvar_units = "PSU"
    elif sbs_var in {"sal11"}:
        kvar_name = "psal1"
        kvar_units = "PSU"
    elif sbs_var in {"secS-priS"}:
        kvar_name = "psaldiff"
        kvar_units = "PSU"

    # DENSITY
    # sg: potential density, based on greek letter sg
    # sgth: potential denisty sg calculated with potential temperature
    # theta (th)
    elif sbs_var in {"density00"}:
        kvar_name = "rho0"
        kvar_units = "density, kg/m^3"
    elif sbs_var in {"sigma-�00"}:
        kvar_name = "sgth0"
        kvar_units = "sigma-theta (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-t00"}:
        kvar_name = "sgt0"
        kvar_units = "sigma-t (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-100"}:
        kvar_name = "sg10"
        kvar_units = "sigma-1 (p=1000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-200"}:
        kvar_name = "sg20"
        kvar_units = "sigma-2 (p=2000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-300"}:
        kvar_name = "sg30"
        kvar_units = "sigma-3 (p=3000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-400"}:
        kvar_name = "sg40"
        kvar_units = "sigma-4 (p=4000 db), kg/m^3-1000"
    elif sbs_var in {"density11"}:
        kvar_name = "rho1"
        kvar_units = "density, kg/m^3"
    elif sbs_var in {"sigma-�11"}:
        kvar_name = "sgth1"
        kvar_units = "sigma-theta (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-t11"}:
        kvar_name = "sgt1"
        kvar_units = "sigma-t (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-111"}:
        kvar_name = "sg11"
        kvar_units = "sigma-1 (p=1000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-211"}:
        kvar_name = "sg21"
        kvar_units = "sigma-2 (p=2000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-311"}:
        kvar_name = "sg31"
        kvar_units = "sigma-3 (p=3000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-411"}:
        kvar_name = "sg41"
        kvar_units = "sigma-4 (p=4000 db), kg/m^3-1000"

    # DENSITY DIFFERENCE
    elif sbs_var in {"D2-D1, d"}:
        kvar_name = "rhodiff"
        kvar_units = "density, kg/m^3"
    elif sbs_var in {"D2-D1,th"}:
        kvar_name = "sgthdiff"
        kvar_units = "sigma-theta, kg/m^3"
    elif sbs_var in {"D2-D1,t"}:
        kvar_name = "sgtdiff"
        kvar_units = "sigma-t, kg/m^3"
    elif sbs_var in {"D2-D1,1"}:
        kvar_name = "sg1diff"
        kvar_units = "sigma-1 (p=1000 db), kg/m^3"
    elif sbs_var in {"D2-D1,2"}:
        kvar_name = "sg2diff"
        kvar_units = "sigma-2 (p=1000 db), kg/m^3"
    elif sbs_var in {"D2-D1,3"}:
        kvar_name = "sg3diff"
        kvar_units = "sigma-3 (p=1000 db), kg/m^3"
    elif sbs_var in {"D2-D1,4"}:
        kvar_name = "sg4diff"
        kvar_units = "sigma-4 (p=1000 db), kg/m^3"

    # SOUND VELOCITY
    # primary sensor
    elif sbs_var in {"svCM"}:
        kvar_name = "sv0CM"
        kvar_units = "Chen-Millero, m/s"
    elif sbs_var in {"svCF"}:
        kvar_name = "sv0CF"
        kvar_units = "Chen-Millero, ft/s"
    elif sbs_var in {"svDM"}:
        kvar_name = "sv0DM"
        kvar_units = "Delgrosso, m/s"
    elif sbs_var in {"svDF"}:
        kvar_name = "sv0DF"
        kvar_units = "Delgrosso, ft/s"
    elif sbs_var in {"svWM"}:
        kvar_name = "sv0WM"
        kvar_units = "Wilson, m/s"
    elif sbs_var in {"svWF"}:
        kvar_name = "sv0WF"
        kvar_units = "Wilson, ft/s"
    # secondary sensor
    elif sbs_var in {"svCM1"}:
        kvar_name = "sv1CM"
        kvar_units = "Chen-Millero, m/s"
    elif sbs_var in {"svCF1"}:
        kvar_name = "sv1CF"
        kvar_units = "Chen-Millero, ft/s"
    elif sbs_var in {"svDM1"}:
        kvar_name = "sv1DM"
        kvar_units = "Delgrosso, m/s"
    elif sbs_var in {"svDF1"}:
        kvar_name = "sv1DF"
        kvar_units = "Delgrosso, ft/s"
    elif sbs_var in {"svWM1"}:
        kvar_name = "sv1WM"
        kvar_units = "Wilson, m/s"
    elif sbs_var in {"svWF1"}:
        kvar_name = "sv1WF"
        kvar_units = "Wilson, ft/s"
    # IOW sound velocity sensor
    elif sbs_var in {"iowSv"}:
        kvar_name = "iowSv"
        kvar_units = "IOW sound velocity sensor, m/s"
    elif sbs_var in {"sbeSv-iowSv"}:
        kvar_name = "svdiff"
        kvar_units = "SBE CTD - IOW SV sensor, m/s"
    # AVERAGE SOUND VELOCITY
    elif sbs_var in {"avgsvCM"}:
        kvar_name = "avgsvCM"
        kvar_units = "Chen-Millero, m/s"
    elif sbs_var in {"avgsvCF"}:
        kvar_name = "avgsvCF"
        kvar_units = "Chen-Millero, ft/s"
    elif sbs_var in {"avgsvDM"}:
        kvar_name = "avgsvDM"
        kvar_units = "Delgrosso, m/s"
    elif sbs_var in {"avgsvDF"}:
        kvar_name = "avgsvDF"
        kvar_units = "Delgrosso, ft/s"
    elif sbs_var in {"avgsvWM"}:
        kvar_name = "avgsvWM"
        kvar_units = "Wilson, m/s"
    elif sbs_var in {"avgsvWF"}:
        kvar_name = "avgsvWF"
        kvar_units = "Wilson, ft/s"

    # BUOYANCY
    elif sbs_var in {"N"}:
        kvar_name = "N"
        kvar_units = "cycles/hour"
    elif sbs_var in {"N^2"}:
        kvar_name = "N2"
        kvar_units = "rad^2/s^2"

    # ACCELERATION
    elif sbs_var in {"accM"}:
        kvar_name = "accM"
        kvar_units = "m/s^2"
    elif sbs_var in {"accF"}:
        kvar_name = "accF"
        kvar_units = "ft/s^2"

    # DESCENT RATE
    elif sbs_var in {"dz/dtM"}:
        kvar_name = "dzdtM"
        kvar_units = "m/s"
    elif sbs_var in {"dz/dtF"}:
        kvar_name = "dzdtF"
        kvar_units = "ft/s"

    # SEAFLOOR DEPTH
    elif sbs_var in {"sfdSM"}:
        kvar_name = "sfdSM"
        kvar_units = "salt water, m"
    elif sbs_var in {"sfdSF"}:
        kvar_name = "sfdSF"
        kvar_units = "salt water, ft"
    elif sbs_var in {"sfdFM"}:
        kvar_name = "sfdFM"
        kvar_units = "fresh water, m"
    elif sbs_var in {"sfdFF"}:
        kvar_name = "sfdFF"
        kvar_units = "fresh water, ft"

    # OTHER DERIVED VARIABLES
    # DYNAMIC METERS
    elif sbs_var in {"dm"}:
        kvar_name = "dm"
        kvar_units = "10 J/kg"
    # GEOPOTENTIAL ANOMALY
    elif sbs_var in {"gpa"}:
        kvar_name = "gpa"
        kvar_units = "J/kg"
    # PLUME ANOMALY
    elif sbs_var in {"pla"}:
        kvar_name = "pla"
        kvar_units = "?"
    # SPECIFIC VOLUME ANOMALY
    elif sbs_var in {"sva"}:
        kvar_name = "sva"
        kvar_units = "10^-8 *m^3/kg"
    # STABILITY
    elif sbs_var in {"E"}:
        kvar_name = "E"
        kvar_units = "rad^2/m"
    elif sbs_var in {"E10^-8"}:
        kvar_name = "E10e_8"
        kvar_units = "10^-8 *rad^2/m"
    # THERMOSTERIC ANOMALY
    elif sbs_var in {"tsa"}:
        kvar_name = "tsa"
        kvar_units = "10^-8 *m^3/kg"
    # SPECIFIC CONDUCTANCE
    elif sbs_var in {"specc"}:
        kvar_name = "specc"
        kvar_units = "uS/cm"
    elif sbs_var in {"speccumhoscm"}:
        kvar_name = "speccumhoscm"
        kvar_units = "umhos/cm"
    elif sbs_var in {"speccmsm"}:
        kvar_name = "speccmsm"
        kvar_units = "mS/cm"
    elif sbs_var in {"speccmmhoscm"}:
        kvar_name = "speccmmhoscm"
        kvar_units = "mmhos/cm"

    # BIOGEOCHEMICAL SENSORS

    # OXYGEN
    # RAW OXYGEN, SBE 43
    elif sbs_var in {"sbeox0V"}:
        kvar_name = "sbeox0V"
        kvar_units = "V"
    elif sbs_var in {"sbeox0F"}:
        kvar_name = "sbeox0F"
        kvar_units = "Hz"
    elif sbs_var in {"sbeox1V"}:
        kvar_name = "sbeox1V"
        kvar_units = "V"
    elif sbs_var in {"sbeox1F"}:
        kvar_name = "sbeox1F"
        kvar_units = "Hz"
    # DERIVED OXYGEN, SBE 43
    elif sbs_var in {"sbeox0ML/L"}:
        kvar_name = "sbeox0mL_L"
        kvar_units = "mL/L"
    elif sbs_var in {"sbeox0Mg/L"}:
        kvar_name = "sbeox0mg_L"
        kvar_units = "mg/L"
    elif sbs_var in {"sbeox0PS"}:
        kvar_name = "sbeox0PS"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeox0Mm/Kg"}:
        kvar_name = "sbeox0mum_kg"
        kvar_units = "micromol/kg"
    elif sbs_var in {"sbeox0Mm/L"}:
        kvar_name = "sbeox0mum_L"
        kvar_units = "micromole/L"
    elif sbs_var in {"sbeox0dOV/dT"}:
        kvar_name = "sbeox0dOV_dT"
        kvar_units = "V/s"
    elif sbs_var in {"sbeox1ML/L"}:
        kvar_name = "sbeox1mL_L"
        kvar_units = "mL/L"
    elif sbs_var in {"sbeox1Mg/L"}:
        kvar_name = "sbeox1mg_L"
        kvar_units = "mg/L"
    elif sbs_var in {"sbeox1PS"}:
        kvar_name = "sbeox1PS"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeox1Mm/Kg"}:
        kvar_name = "sbeox1mum_kg"
        kvar_units = "micromol/kg"
    elif sbs_var in {"sbeox1Mm/L"}:
        kvar_name = "sbeox1mum_L"
        kvar_units = "micromole/L"
    elif sbs_var in {"sbeox1dOV/dT"}:
        kvar_name = "sbeox1dOV_dT"
        kvar_units = "V/s"
    # Difference between SBE 43 sensors
    elif sbs_var in {"sbeox0ML/Ldiff"}:
        kvar_name = "sbeox0ML_Ldiff"
        kvar_units = "ml/l"
    elif sbs_var in {"sbeox0Mg/Ldiff"}:
        kvar_name = "sbeox0Mg_Ldiff"
        kvar_units = "mg/l"
    elif sbs_var in {"sbeox0PSdiff"}:
        kvar_name = "sbeox0PSdiff"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeox0Mm/Kgdiff"}:
        kvar_name = "sbeox0Mm_Kgdiff"
        kvar_units = "umol/kg"
    elif sbs_var in {"sbeox0Mm/Ldiff"}:
        kvar_name = "sbeox0Mm_Ldiff"
        kvar_units = "umol/l"
    # RAW OXYGEN, SBE 63
    elif sbs_var in {"sbeoxpd"}:
        kvar_name = "sbeoxpd"
        kvar_units = "usec"
    elif sbs_var in {"sbeoxpdv"}:
        kvar_name = "sbeoxpdv"
        kvar_units = "V"
    elif sbs_var in {"sbeoxpd1"}:
        kvar_name = "sbeoxpd1"
        kvar_units = "usec"
    elif sbs_var in {"sbeoxpdv1"}:
        kvar_name = "sbeoxpdv1"
        kvar_units = "sbeoxpdv1"
    elif sbs_var in {"sbeoxtv"}:
        kvar_name = "sbeoxtv"
        kvar_units = "sbeoxtv"
    elif sbs_var in {"sbeoxtv1"}:
        kvar_name = ""
        kvar_units = "sbeoxtv1"
    # RAW OXYGEN TEMPERATURE, SBE 63
    elif sbs_var in {"sbeoxTC"}:
        kvar_name = "sbeoxTC"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"sbeoxTF"}:
        kvar_name = "sbeoxTF"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"sbeoxTC1"}:
        kvar_name = "sbeoxTC1"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"sbeoxTF1"}:
        kvar_name = "sbeoxTF1"
        kvar_units = "ITS-90, deg F"
    # DERIVED OXYGEN, SBE 63
    elif sbs_var in {"sbeopoxML/L"}:
        kvar_name = "sbeopoxML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"sbeopoxMg/L"}:
        kvar_name = "sbeopoxMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"sbeopoxPS"}:
        kvar_name = "sbeopoxPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeopoxMm/Kg"}:
        kvar_name = "sbeopoxMm_Kg"
        kvar_units = "umol/kg"
    elif sbs_var in {"sbeopoxMm/L"}:
        kvar_name = "sbeopoxMm_L"
        kvar_units = "umol/l"
    elif sbs_var in {"sbeopoxML/L1"}:
        kvar_name = "sbeopoxML_L1"
        kvar_units = "ml/l"
    elif sbs_var in {"sbeopoxMg/L1"}:
        kvar_name = "sbeopoxMg_L1"
        kvar_units = "mg/l"
    elif sbs_var in {"sbeopoxPS1"}:
        kvar_name = "sbeopoxPS1"
        kvar_units = r"% saturation"
    elif sbs_var in {"SbeopoxMm/Kg1"}:
        kvar_name = "SbeopoxMm_Kg1"
        kvar_units = "umol/kg"
    elif sbs_var in {"sbeopoxMm/L1"}:
        kvar_name = "sbeopoxMm/L1"
        kvar_units = "umol/l"
    # AANDERAA OPTODE
    elif sbs_var in {"opoxML/L"}:
        kvar_name = "opoxML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"opoxMg/L"}:
        kvar_name = "opoxMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"opoxPS"}:
        kvar_name = "opoxPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"opoxMm/L"}:
        kvar_name = "opoxMm_L"
        kvar_units = "umol/l"
    # BECKMAN/YSI
    elif sbs_var in {"oxC"}:
        kvar_name = "oxC"
        kvar_units = "uA"
    elif sbs_var in {"oxsC"}:
        kvar_name = "oxsC"
        kvar_units = "uA"
    elif sbs_var in {"oxTC"}:
        kvar_name = "oxTC"
        kvar_units = "deg C"
    elif sbs_var in {"oxTF"}:
        kvar_name = "oxTF"
        kvar_units = "deg F"
    elif sbs_var in {"oxsTC"}:
        kvar_name = "oxsTC"
        kvar_units = "deg C"
    elif sbs_var in {"oxsTF"}:
        kvar_name = "oxsTF"
        kvar_units = "deg F"
    elif sbs_var in {"oxML/L"}:
        kvar_name = "oxML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxMg/L"}:
        kvar_name = "oxMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxPS"}:
        kvar_name = "oxPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"oxMm/Kg"}:
        kvar_name = "oxMm_Kg"
        kvar_units = "umol/kg"
    elif sbs_var in {"oxdOC/dT"}:
        kvar_name = "oxdOC_dT"
        kvar_units = "doc/dt"
    elif sbs_var in {"oxsML/L"}:
        kvar_name = "oxsML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxsMg/L"}:
        kvar_name = "oxsMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxsPS"}:
        kvar_name = "oxsPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"oxsMm/Kg"}:
        kvar_name = "oxsMm_Kg"
        kvar_units = "umol/kg"
    elif sbs_var in {"oxsdOC/dT"}:
        kvar_name = "oxsdOC_dT"
        kvar_units = "doc/dt"
    # IOW OXYGEN SATURATIONS
    elif sbs_var in {"iowOxML/L"}:
        kvar_name = "iowOxML/L"
        kvar_units = "ml/l"
    # GARCIA & GORDON OXYGEN SATURATION
    elif sbs_var in {"oxsolML/L"}:
        kvar_name = "oxsolML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxsolMg/L"}:
        kvar_name = "oxsolMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxsolMm/Kg"}:
        kvar_name = "oxsolMm_Kg"
        kvar_units = "umol/kg"
    # WEISS OXYGEN SATURATION
    elif sbs_var in {"oxsatML/L"}:
        kvar_name = "oxsatML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxsatMg/L"}:
        kvar_name = "oxsatMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxsatMm/Kg"}:
        kvar_name = "oxsatMm_Kg"
        kvar_units = "umol/kg"

    # OPTICAL SENSORS
    # TURNER CYCLOPS
    # CDOM
    elif sbs_var in {"cdomflTC0", "cdomflTC1", "cdomflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb QS"
    # FLUORESCENCE
    elif sbs_var in {"chloroflTC0", "chloroflTC1", "chloroflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ug/l"
    # CRUDE OIL
    elif sbs_var in {"croilflTC0", "croilflTC1", "croilflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb QS"
    # FLUORESCEIN
    elif sbs_var in {"flflTC0", "flflTC1", "flflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb"
    # OPTICAL BRIGHTENERS
    elif sbs_var in {"obrflTC0", "obrflTC1", "obrflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb QS"
    # PHYCOCYANIN
    elif sbs_var in {"phycyflTC0", "phycyflTC1", "phycyflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "RFU"
    # PHYCOERYTHRIN
    elif sbs_var in {"phyeryflTC0", "phyeryflTC1", "phyeryflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "RFU"
    # REFINED FUELS
    elif sbs_var in {"rfuels0", "rfuels1", "rfuelsdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb NS"
    # RHODAMINE
    elif sbs_var in {"rhodflTC0", "rhodflTC1", "rhodflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb"
    # TURNER CYCLOPS TURBIDITY
    elif sbs_var in {"turbflTC0", "turbflTC1", "turbflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "NTU"

    # BIOSPHERICAL
    # FLUORESCENCE
    elif sbs_var in {"chConctr", "naFluor", "product"}:
        kvar_name = sbs_var
        kvar_units = ""
    # CHELSEA AQUA 3
    # FLUORESCENCE
    elif sbs_var in {"flC", "flC1", "flCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ug/l"
    # CHELSEA MINI CHL CON
    # FLUORESCENCE
    elif sbs_var in {"flCM"}:
        kvar_name = "flCM"
        kvar_units = "ug/l"
    # CHELSEA UV AQUATRACKA
    # FLUORESCENCE
    elif sbs_var in {"flCUVA", "flCUVA1", "flCUVAdiff"}:
        kvar_name = sbs_var
        kvar_units = "ug/l"
    # DR HAARDT
    # Fluorescence: Chlorophyll a, Phycoerythrin, Yellow Sub
    elif sbs_var in {"haardtC", "haardtP", "haardtY"}:
        kvar_name = sbs_var
        kvar_units = ""
    # SEAPOINT
    # FLUORESCENCE
    elif sbs_var in {"flSP", "flSP1", "flSPdiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # RHODAMINE
    elif sbs_var in {"flSPR"}:
        kvar_name = sbs_var
        kvar_units = ""
    # ULTRAVIOLET
    elif sbs_var in {"flSPuv0", "flSPuv1", "flSPuvdiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # SEATECH FLUORESCENCE
    elif sbs_var in {"flS"}:
        kvar_name = sbs_var
        kvar_units = ""
    # TURNER 10-005
    elif sbs_var in {"flT"}:
        kvar_name = "flT"
        kvar_units = ""
    # TURNER 10-Au-005
    elif sbs_var in {"flTAu"}:
        kvar_name = "flTAu"
        kvar_units = ""
    # TURNER SCUFA CORRECTED
    elif sbs_var in {"flSCC", "flSCC1", "flSCCdiff"}:
        kvar_name = sbs_var
        kvar_units = "RFU"
    # TURNER SCUFA
    elif sbs_var in {"flScufa", "flScufa1", "flScufadiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # WETLABS AC3
    elif sbs_var in {"wetChAbs"}:
        kvar_name = "wetChAbs"
        kvar_units = "1/m"
    # WETLABS CDOM
    elif sbs_var in {
        "wetCDOM",
        "wetCDOM1",
        "wetCDOM2",
        "wetCDOM3",
        "wetCDOM4",
        "wetCDOM5",
        "wetCDOMdiff",
    }:
        kvar_name = sbs_var
        kvar_units = "mg/m^3"
    # WETLABS AC3 CHLOROPHYLL
    elif sbs_var in {"wetChConc"}:
        kvar_name = "wetChConc"
        kvar_units = "mg/m^3"
    # WETLABS ECO-AFL
    elif sbs_var in {"flECO-AFL"}:
        kvar_name = "flECO_AFL"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL1"}:
        kvar_name = "flECO_AFL1"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL2"}:
        kvar_name = "flECO_AFL2"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL3"}:
        kvar_name = "flECO_AFL3"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL4"}:
        kvar_name = "flECO_AFL4"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL5"}:
        kvar_name = "flECO_AFL5"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFLdiff"}:
        kvar_name = "flECO_AFLdiff"
        kvar_units = "mg/m^3"
    # WETLABS SEAOWL
    # FLUOROMETER
    elif sbs_var in {"flWETSeaOWLchl0", "flWETSeaOWLchl1", "flWETSeaOWLchldiff"}:
        kvar_name = sbs_var
        kvar_units = "?g/l"
    # FDOM
    elif sbs_var in {"flWETSeaOWLfdom0", "flWETSeaOWLfdom1", "flWETSeaOWLfdomdiff"}:
        kvar_name = sbs_var
        kvar_units = "?g/l"
    # WETSTAR
    elif sbs_var in {
        "wetStar",
        "wetStar1",
        "wetStar2",
        "wetStar3",
        "wetStar4",
        "wetStar5",
        "wetStardiff",
    }:
        kvar_name = sbs_var
        kvar_units = "mg/m^3"

    # BACKSCATTER
    # D & A BACKSCATTERANCE
    elif sbs_var in {"obs", "obs1", "obsdiff"}:
        kvar_name = sbs_var
        kvar_units = "NTU"
    # CHELSEA NEPHELOMETER
    elif sbs_var in {"nephc"}:
        kvar_name = "nephc"
        kvar_units = "FTU"
    # D & A
    elif sbs_var in {"obs3+"}:
        kvar_name = "obs3plus"
        kvar_units = "NTU"
    elif sbs_var in {"obs3+1"}:
        kvar_name = "obs3plus1"
        kvar_units = "NTU"
    elif sbs_var in {"obs3+diff"}:
        kvar_name = "obs3plusdiff"
        kvar_units = "NTU"
    # DR. HAARDT
    elif sbs_var in {"haardtT"}:
        kvar_name = "haardtT"
        kvar_units = ""
    # IFREMER
    elif sbs_var in {"diff"}:
        kvar_name = "diff"
        kvar_units = ""
    # SEATECH LS6000
    elif sbs_var in {"stLs6000", "stLs60001", "stLs6000diff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # TURNER SCUFA
    elif sbs_var in {"obsscufa", "obsscufa1", "obsscufadiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # WETLABS RAW COUNTS
    elif sbs_var in {"wl0", "wl1", "wl2", "wl3", "wl4", "wl5"}:
        kvar_name = sbs_var
        kvar_units = "Counts"

    # TURBIDITY
    # SEAPOINT TURBIDITY
    elif sbs_var in {"seaTurbMtr", "seaTurbMtr1", "seaTurbMtrdiff"}:
        kvar_name = sbs_var
        kvar_units = "FTU"
    # WETLABS ECO BB
    elif sbs_var in {
        "turbWETbb0",
        "turbWETbb1",
        "turbWETbb2",
        "turbWETbb3",
        "turbWETbb4",
        "turbWETbb5",
        "turbWETbbdiff",
    }:
        kvar_name = sbs_var
        kvar_units = "m^-1/sr"
    # WETLABS ECO
    elif sbs_var in {
        "turbWETntu0",
        "turbWETntu1",
        "turbWETntu2",
        "turbWETntu3",
        "turbWETntu4",
        "turbWETntu5",
        "turbWETntudiff",
    }:
        kvar_name = sbs_var
        kvar_units = "NTU"
    # WETLABS SEAOWL
    elif sbs_var in {"turbWETSeaOWLbb0", "turbWETSeaOWLbb1", "turbWETSeaOWLbbdiff"}:
        kvar_name = sbs_var
        kvar_units = "m^-1/sr"

    # TRANSMISSOMETERS
    # CHELSEA/SEATECH BEAM ATTENUATION
    elif sbs_var in {"bat", "bat1", "batdiff"}:
        kvar_name = sbs_var
        kvar_units = "1/m"
    # WETLABS AC3 BEAM TRANSMISSION
    elif sbs_var in {"wetBAttn"}:
        kvar_name = "wetBAttn"
        kvar_units = "1/m"
    # WETLABS C-STAR BEAM ATTENUATION
    elif sbs_var in {
        "CStarAt",
        "CStarAt0",
        "CStarAt1",
        "CStarAt2",
        "CStarAt3",
        "CStarAt4",
        "CStarAt5",
        "CStarAtdiff",
    }:
        kvar_name = sbs_var
        kvar_units = "1/m"
    # CHELSEA/SEATECH BEAM TRANSMISSION
    elif sbs_var in {"xmiss", "xmiss1", "xmissdiff"}:
        kvar_name = sbs_var
        kvar_units = r"%"
    # WETLABS AC3 BEAM TRANSMISSION
    elif sbs_var in {"wetBTrans"}:
        kvar_name = "wetBTrans"
        kvar_units = r"%"
    # WETLABS C-STAR BEAM TRANSMISSION
    elif sbs_var in {
        "CStarTr",
        "CStarTr0",
        "CStarTr1",
        "CStarTr2",
        "CStarTr3",
        "CStarTr4",
        "CStarTr5",
        "CStarTrdiff",
    }:
        kvar_name = sbs_var
        kvar_units = r"%"

    # LISST
    elif sbs_var in {"lisstBC"}:
        kvar_name = "lisstBC"
        kvar_units = "1/m"
    elif sbs_var in {"lisstOT"}:
        kvar_name = "lisstOT"
        kvar_units = r"%"
    elif sbs_var in {"lisstMD"}:
        kvar_name = "lisstMD"
        kvar_units = "u"
    elif sbs_var in {"lisstTVC"}:
        kvar_name = "lisstTVC"
        kvar_units = "ul/l"

    # PAR
    elif sbs_var in {"cpar"}:
        kvar_name = "cpar"
        kvar_units = r"%"
    elif sbs_var in {"par"}:
        kvar_name = "par"
        kvar_units = ""
    elif sbs_var in {"par1"}:
        kvar_name = "par1"
        kvar_units = ""
    elif sbs_var in {"par/log"}:
        kvar_name = "par_log"
        kvar_units = "umol photons/m2/s"
    elif sbs_var in {"spar"}:
        kvar_name = "spar"
        kvar_units = ""

    # BIOGEOCHEMICAL SENSORS
    # METHANE FRANATECH METS
    elif sbs_var in {"meth"}:
        kvar_name = "meth"
        kvar_units = "umol/l"
    elif sbs_var in {"methT"}:
        kvar_name = "methT"
        kvar_units = "deg C"
    # NITROGEN
    elif sbs_var in {"n2satML/L"}:
        kvar_name = "n2satML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"n2satMg/L"}:
        kvar_name = "n2satMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"n2satumol/kg"}:
        kvar_name = "n2satumol_kg"
        kvar_units = "umol/kg"
    # OXIDATION REDUCTION POTENTIAL
    elif sbs_var in {"orp"}:
        kvar_name = "orp"
        kvar_units = "mV"
    # pH
    elif sbs_var in {"ph", "phInt", "phExt"}:
        kvar_name = sbs_var
        kvar_units = ""
    # ZAPS
    elif sbs_var in {"zaps"}:
        kvar_name = "zaps"
        kvar_units = "nmol"

    # USER DEFINED VARIABLE
    elif sbs_var in {"user", "user1", "user2", "user3", "user4", "user5"}:
        kvar_name = sbs_var
        kvar_units = ""
    # USER POLYNOMIAL
    elif sbs_var in {"upoly0", "upoly1", "upoly2"}:
        kvar_name = sbs_var
        kvar_units = ""

    # PACKAGE STATUS
    # ALTIMETER
    elif sbs_var in {"altM"}:
        kvar_name = "altM"
        kvar_units = "m"
    elif sbs_var in {"altF"}:
        kvar_name = "altF"
        kvar_units = "ft"

        # Bottom Contact
    elif sbs_var in {"bct"}:
        kvar_name = "bct"
        kvar_units = ""

    # WATER SAMPLER STATUS
    # BOTTLE POSITION, BOTTLES CLOSED (HB), BOTTLES FIRED, NEW POSITION'
    elif sbs_var in {"bpos", "HBBotCls", "nbf", "newpos"}:
        kvar_name = sbs_var
        kvar_units = ""
    # PUMP STATUS
    elif sbs_var in {"pumps"}:
        kvar_name = "pumps"
        kvar_units = ""

    # DATA STREAM
    # BYTE COUNT
    elif sbs_var in {"nbytes"}:
        kvar_name = "nbytes"
        kvar_units = ""
    # MODULO ERROR COUNT
    elif sbs_var in {"modError"}:
        kvar_name = "modError"
        kvar_units = ""
    # MODULO WORD
    elif sbs_var in {"mod"}:
        kvar_name = "mod"
        kvar_format = r"%s"
        kvar_units = ""
    # SCAN COUNT
    elif sbs_var in {"scan"}:
        kvar_name = "scan"
        kvar_units = ""
    # SCANS PER BIN
    elif sbs_var in {"nbin"}:
        kvar_name = "nbin"
        kvar_units = ""

    # FLAGS
    elif sbs_var in {"flag"}:
        kvar_name = "flag"
        kvar_units = ""

    else:
        # assign the original SBS name to the output variable
        kvar_name = sbs_var
        # find and replace all the characters that are not alphabetic,
        # numbers or underscores with underscores
        kvar_name = re.sub(r"\W", "_", kvar_name)
        # print(f"Assigning {sbs_var} as {kvar_name}")
        kvar_units = ""

    #     elif sbs_var in  ''
    #         kvar_name = ''
    # #         kvar_units = ''

    return {"name": kvar_name, "format": kvar_format, "units": kvar_units}



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""A collection of raw data conversion functions."""

# Native imports
from math import e, floor
from typing import Literal
import warnings

# Third-party imports
import gsw
import numpy as np
from numpy.polynomial import Polynomial
from scipy import stats

# Sea-Bird imports

# Internal imports
import seabirdscientific.cal_coefficients as cc


DBAR_TO_PSI = 1.450377
PSI_TO_DBAR = 0.6894759
OXYGEN_PHASE_TO_VOLTS = 39.457071
KELVIN_OFFSET_0C = 273.15
KELVIN_OFFSET_25C = 298.15
OXYGEN_MLPERL_TO_MGPERL = 1.42903
OXYGEN_MLPERL_TO_UMOLPERKG = 44660
# taken from https://blog.seabird.com/ufaqs/what-is-the-difference-in-temperature-expressions-between-ipts-68-and-its-90/
ITS90_TO_IPTS68 = 1.00024
# micro moles of nitrate to milligrams of nitrogen per liter
UMNO3_TO_MGNL = 0.014007
# [J K^{-1} mol^{-1}] Gas constant from SBS application note 99
R = 8.3144621
# [Coulombs mol^{-1}] Faraday constant from SBS application note 99
F = 96485.365


def convert_temperature(
    temperature_counts_in: np.ndarray,
    coefs: cc.TemperatureCoefficients,
    standard: Literal["ITS90", "IPTS68"] = "ITS90",
    units: Literal["C", "F"] = "C",
    use_mv_r: bool = False,
):
    """Returns the value after converting it to degrees C, ITS-90.

    Data is expected to be raw data from an instrument in A/D counts

    :param temperature_counts_in: temperature value to convert in A/D
        counts
    :param coefs: calibration coefficients for the temperature sensor
    :param standard: whether to use ITS90 or to use IPTS-68 calibration
        standard
    :param units: whether to use celsius or to convert to fahrenheit
    :param use_mv_r: true to perform extra conversion steps required by
        some instruments (check the cal sheet to see if this is required)

    :return: temperature val converted to ITS-90 or IPTS68 in degrees C or F
    """

    if use_mv_r:
        mv = (temperature_counts_in - 524288) / 1.6e007
        r = (mv * 2.900e009 + 1.024e008) / (2.048e004 - mv * 2.0e005)
        temperature_counts = r
    else:
        temperature_counts = temperature_counts_in

    log_t = np.log(temperature_counts)
    temperature = (
        1 / (coefs.a0 + coefs.a1 * log_t + coefs.a2 * log_t**2 + coefs.a3 * log_t**3)
    ) - KELVIN_OFFSET_0C

    if standard == "IPTS68":
        temperature *= ITS90_TO_IPTS68
    if units == "F":
        temperature = temperature * 9 / 5 + 32  # Convert C to F

    return temperature


def convert_temperature_frequency(
    frequency: np.ndarray,
    coefs: cc.TemperatureFrequencyCoefficients,
    standard: Literal["ITS90", "IPTS68"] = "ITS90",
    units: Literal["C", "F"] = "C",
):
    """Convert raw frequency to temperature in degrees Celsius or degrees Fahrenheit

    :param frequency: raw frequency from the temperature sensor
    :param coefs: calibration coefficients for the temperature sensor
    :return: temperature in Celsius or Fahrenheit
    """
    fLog = np.log(coefs.f0 / frequency)
    temperature = (
        1 / (coefs.g + coefs.h * fLog + coefs.i * fLog**2 + coefs.j * fLog**3) - KELVIN_OFFSET_0C
    )

    if standard == "IPTS68":
        temperature *= ITS90_TO_IPTS68
    if units == "F":
        temperature = temperature * 9 / 5 + 32  # Convert C to F

    return temperature


def convert_pressure(
    pressure_count: np.ndarray,
    compensation_voltage: np.ndarray,
    coefs: cc.PressureCoefficients,
    units: Literal["dbar", "psia", "psig"] = "psig",
):
    """Converts pressure counts to sea pressure (psig and dbar) and absolute pressure (psia)

    pressure_count and compensation_voltage are expected to be raw data
    from an instrument in A/D counts

    :param pressure_count: pressure value to convert, in A/D counts
    :param compensation_voltage: pressure temperature compensation
        voltage, in counts or volts depending on the instrument
    :param coefs: calibration coefficients for the pressure sensor
    :param units: whether or not to use psig or dbar as the returned
        unit type

    :return: sea pressure val in dbar or PSIG
    """
    sea_level_pressure = 14.7

    t = (
        coefs.ptempa0
        + coefs.ptempa1 * compensation_voltage
        + coefs.ptempa2 * compensation_voltage**2
    )
    x = pressure_count - coefs.ptca0 - coefs.ptca1 * t - coefs.ptca2 * t**2
    n = x * coefs.ptcb0 / (coefs.ptcb0 + coefs.ptcb1 * t + coefs.ptcb2 * t**2)
    pressure = coefs.pa0 + coefs.pa1 * n + coefs.pa2 * n**2

    if units == "dbar" or units == "psig":
        pressure -= sea_level_pressure

    if units == "dbar":
        pressure *= PSI_TO_DBAR

    return pressure


def convert_pressure_digiquartz(
    pressure_count: np.ndarray,
    compensation_voltage: np.ndarray,
    coefs: cc.PressureDigiquartzCoefficients,
    units: Literal["dbar", "psia"],
    sample_interval: float,
):
    """Converts pressure counts to PSIA (pounds per square inch, abolute) or dbar for a digiquartz pressure sensor.

    pressure_count and compensation_voltage are expected to be raw data
    from an instrument in A/D counts

    :param pressure_count: pressure value to convert, in A/D counts
    :param compensation_voltage: pressure temperature compensation
        voltage, in counts or volts depending on the instrument
    :param coefs: calibration coefficients for the digiquartz pressure sensor
    :param units: whether or not to use psia or dbar as the returned
        unit type
    :param sample_interval: sample rate of the data to be used for temperature compensation correction, in seconds
    :return: pressure val in PSIA or dbar
    """
    sea_level_pressure = 14.7
    # First, average temperature compensation over 30 seconds
    max_scans_in_30_seconds = 720
    scans_in_window = floor(30 / sample_interval)
    scans_in_window = max(scans_in_window, 1)
    scans_in_window = min(scans_in_window, max_scans_in_30_seconds)

    rolling_sum = compensation_voltage[0] * scans_in_window
    modified_compensation_voltage = compensation_voltage.copy()

    for i in range(0, len(compensation_voltage)):
        if i < scans_in_window:
            # remove a copy of 0-index value from rolling sum
            rolling_sum -= compensation_voltage[0]
        else:
            # remove oldest value from rolling sum
            rolling_sum -= compensation_voltage[i - scans_in_window]

        rolling_sum += compensation_voltage[i]
        modified_compensation_voltage[i] = (
            rolling_sum / scans_in_window * coefs.AD590M + coefs.AD590B
        )

    # Now, calculate pressure

    t = 1 / pressure_count * 1000000  # convert to period in usec
    c = (
        coefs.c1
        + coefs.c2 * modified_compensation_voltage
        + coefs.c3 * modified_compensation_voltage**2
    )
    d = coefs.d1 + coefs.d2 * modified_compensation_voltage
    t0 = (
        coefs.t1
        + coefs.t2 * modified_compensation_voltage
        + coefs.t3 * modified_compensation_voltage**2
        + coefs.t4 * modified_compensation_voltage**3
        + coefs.t5 * modified_compensation_voltage**4
    )

    t0_squared_over_t_squared = (t0**2) / (t**2)
    one_minus_ratio = 1 - t0_squared_over_t_squared
    p = c * one_minus_ratio * (1 - d * one_minus_ratio)
    abs_pressure = p - sea_level_pressure
    if units == "dbar":
        abs_pressure *= PSI_TO_DBAR
    return abs_pressure


def convert_conductivity(
    conductivity_count: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
    coefs: cc.ConductivityCoefficients,
    scalar: float = 1.0,
):
    """Converts raw conductivity counts to S/m.

    Data is expected to be raw data from instrument in A/D counts

    :param conductivity_count: conductivity value to convert, in A/D
        counts
    :param temperature: reference temperature, in degrees C
    :param pressure: reference pressure, in dbar
    :param coefs: calibration coefficient for the conductivity sensor
    :param scalar: value to multiply by at the end. For most instruments, this is 1. For SBE911, it is 1/10

    :return: conductivity val converted to S/m
    """
    f = conductivity_count * np.sqrt(1 + coefs.wbotc * temperature) / 1000
    numerator = coefs.g + coefs.h * f**2 + coefs.i * f**3 + coefs.j * f**4
    denominator = 1 + coefs.ctcor * temperature + coefs.cpcor * pressure
    return numerator / denominator * scalar


def potential_density_from_t_s_p(
    temperature: np.ndarray,
    salinity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
    reference_pressure=0.0,
):
    """Derive potential density from measured temperature, salinity, and
    pressure.

    :param temperature: Measure temperature, in degrees C
    :param salinity: Measured salinity, in practical salinity units
    :param pressure: Measured pressure, in decibars
    :param lon: Longitude
    :param lat: Latitude
    :param reference_pressure: Reference pressure in decibars. Defaults
        to 0.0.

    :return: Potential density in kg/m^3
    """

    absolute_salinity = gsw.SA_from_SP(salinity, pressure, lon, lat)
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature, pressure)
    potential_density = (
        gsw.rho(absolute_salinity, conservative_temperature, reference_pressure) - 1000
    )
    return potential_density


def potential_density_from_t_c_p(
    temperature: np.ndarray,
    conductivity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
    reference_pressure=0.0,
):
    """Derive potential density from measured temperature, salinity, and
    pressure.

    :param temperature: Measure temperature, in degrees C
    :param conductivity: Measured conductivity, in mSiemens/cm
    :param pressure: Measured pressure, in decibars
    :param lon: Longitude
    :param lat: Latitude
    :param reference_pressure: Reference pressure in decibars. Defaults
        to 0.0.

    :return: Potential density in kg/m^3
    """

    salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    return potential_density_from_t_s_p(
        temperature, salinity, pressure, lon, lat, reference_pressure
    )


def density_from_t_s_p(
    temperature: np.ndarray,
    salinity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
):
    """Derive potential density from measured temperature, salinity, and
    pressure.

    :param temperature: Measure temperature, in degrees C
    :param salinity: Measured salinity, in practical salinity units
    :param pressure: Measured pressure, in decibars
    :param lon: Longitude
    :param lat: Latitude

    :return: Potential density in kg/m^3
    """

    absolute_salinity = gsw.SA_from_SP(salinity, pressure, lon, lat)
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature, pressure)
    density = gsw.rho(absolute_salinity, conservative_temperature, pressure)
    return density


def density_from_t_c_p(
    temperature: np.ndarray,
    conductivity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
):
    """Derive potential density from measured temperature, salinity, and
    pressure.

    :param temperature: Measure temperature, in degrees C
    :param conductivity: Measured conductivity, in mSiemens/cm
    :param pressure: Measured pressure, in decibars
    :param lon: Longitude
    :param lat: Latitude

    :return: Potential density in kg/m^3
    """

    salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    return density_from_t_s_p(temperature, salinity, pressure, lon, lat)


def depth_from_pressure(
    pressure_in: np.ndarray,
    latitude: float,
    depth_units: Literal["m", "ft"] = "m",
    pressure_units: Literal["dbar", "psi"] = "dbar",
):
    """Derive depth from pressure and latitude.

    :param pressure: Numpy array of floats representing pressure, in
        dbar or psi
    :param latitude: Latitude (-90.0 to 90.0)
    :param depth_units: 'm' for meters, 'ft' for feet. Defaults to 'm'.
    :param pressure_units: 'dbar' for decibars, 'psi' for PSI. Defaults
        to 'dbar'.

    :return: A numpy array representing depth in meters or feet
    """
    pressure = pressure_in.copy()
    if pressure_units == "psi":
        pressure /= DBAR_TO_PSI

    depth = -gsw.z_from_p(pressure, latitude)

    if depth_units == "ft":
        depth *= 3.28084

    return depth


def convert_sbe63_oxygen(
    raw_oxygen_phase: np.ndarray,
    thermistor: np.ndarray,
    pressure: np.ndarray,
    salinity: np.ndarray,
    coefs: cc.Oxygen63Coefficients,
    thermistor_coefs: cc.Thermistor63Coefficients,
    thermistor_units: Literal["volts", "C"] = "volts",  # Is this volts or frequency?
):
    """Returns the data after converting it to ml/l.

    raw_oxygen_phase is expected to be in raw phase, raw_thermistor_temp
    in counts, pressure in dbar, and salinity in practical salinity (PSU)

    :param raw_oxygen_phase: SBE63 phase value, in microseconds
    :param thermistor_temp: SBE63 thermistor data to use are reference,
        in counts
    :param pressure: Converted pressure value from the attached CTD, in
        dbar
    :param salinity: Converted salinity value from the attached CTD, in
        practical salinity PSU
    :param coefs (cc.Oxygen63Coefficients): calibration coefficients for
        the SBE63 sensor
    :param thermistor_coefs (cc.Thermistor63Coefficients): calibration coefficients for
        the SBE63 thermistor sensor
    :param thermistor_units: units of thermistor_temp input

    :return: converted Oxygen value, in ml/l
    """
    if thermistor_units == "volts":
        temperature = convert_sbe63_thermistor(thermistor, thermistor_coefs)
    elif thermistor_units == "C":
        temperature = thermistor
    else:
        raise ValueError

    oxygen_volts = raw_oxygen_phase / OXYGEN_PHASE_TO_VOLTS  # from the manual

    ksv = coefs.c0 + coefs.c1 * temperature + coefs.c2 * temperature**2

    # The following correction coefficients are all constants
    sol_b0 = -6.24523e-3
    sol_b1 = -7.37614e-3
    sol_b2 = -1.0341e-2
    sol_b3 = -8.17083e-3
    sol_c0 = -4.88682e-7

    ts = np.log((KELVIN_OFFSET_25C - temperature) / (KELVIN_OFFSET_0C + temperature))
    s_corr_exp = (
        salinity * (sol_b0 + sol_b1 * ts + sol_b2 * ts**2 + sol_b3 * ts**3) + sol_c0 * salinity**2
    )
    s_corr = e**s_corr_exp

    # temperature in Kelvin
    temperature_k = temperature + KELVIN_OFFSET_0C
    p_corr_exp = (coefs.e * pressure) / temperature_k
    p_corr = e**p_corr_exp

    # fmt: off
    ox_val = (
        (((coefs.a0 + coefs.a1 * temperature + coefs.a2 * oxygen_volts**2)
        / (coefs.b0 + coefs.b1 * oxygen_volts) - 1.0) / ksv) * s_corr * p_corr
    )
    # fmt: on

    return ox_val


def convert_sbe63_thermistor(
    instrument_output: np.ndarray,
    coefs: cc.Thermistor63Coefficients,
):
    """Converts a SBE63 thermistor raw output array to temperature in
    ITS-90 deg C.

    :param instrument_output: raw values from the thermistor
    :param coefs: calibration coefficients for the thermistor in the
        SBE63 sensor

    :return: converted thermistor temperature values in ITS-90 deg C
    """
    log_raw = np.log((100000 * instrument_output) / (3.3 - instrument_output))
    temperature = (
        1 / (coefs.ta0 + coefs.ta1 * log_raw + coefs.ta2 * log_raw**2 + coefs.ta3 * log_raw**3)
        - KELVIN_OFFSET_0C
    )
    return temperature


def convert_sbe43_oxygen(
    voltage: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
    salinity: np.ndarray,
    coefs: cc.Oxygen43Coefficients,
    apply_tau_correction: bool = False,
    apply_hysteresis_correction: bool = False,
    window_size: float = 1,
    sample_interval: float = 1,
):
    """Returns the data after converting it to ml/l.

    voltage is expected to be in volts, temperature in deg c, pressure
    in dbar, and salinity in practical salinity (PSU). All equation
    information comes from the June 2013 revision of the SBE43 manual

    :param voltage: SBE43 voltage
    :param temperature: temperature value converted to deg C
    :param pressure: Converted pressure value from the attached CTD, in
        dbar
    :param salinity: Converted salinity value from the attached CTD, in
        practical salinity PSU
    :param coefs: calibration coefficients for the SBE43 sensor
    :param apply_tau_correction: whether or not to run tau correction
    :param apply_hysteresis_correction: whether or not to run hysteresis
        correction
    :param window_size: size of the window to use for tau correction, if
        applicable, in seconds
    :param sample_interval: sample rate of the data to be used for tau
        correction, if applicable. In seconds.

    :return: converted Oxygen values, in ml/l
    """
    # start with all 0 for the dvdt
    dvdt_values = np.zeros(len(voltage))
    if apply_tau_correction:
        # Calculates how many scans to have on either side of our median
        # point, accounting for going out of index bounds
        scans_per_side = floor(window_size / 2 / sample_interval)
        for i in range(scans_per_side, len(voltage) - scans_per_side):
            ox_subset = voltage[i - scans_per_side : i + scans_per_side + 1]

            time_subset = np.arange(
                0, len(ox_subset) * sample_interval, sample_interval, dtype=float
            )

            result = stats.linregress(time_subset, ox_subset)

            dvdt_values[i] = result.slope

    correct_ox_voltages = voltage.copy()
    if apply_hysteresis_correction:
        # Hysteresis starts at 1 because 0 can't be corrected
        for i in range(1, len(correct_ox_voltages)):
            # All Equation info from APPLICATION NOTE NO. 64-3
            d = 1 + coefs.h1 * (np.exp(pressure[i] / coefs.h2) - 1)
            c = np.exp(-1 * sample_interval / coefs.h3)
            ox_volts = correct_ox_voltages[i] + coefs.v_offset

            prev_ox_volts_new = correct_ox_voltages[i - 1] + coefs.v_offset
            ox_volts_new = ((ox_volts + prev_ox_volts_new * c * d) - (prev_ox_volts_new * c)) / d
            ox_volts_final = ox_volts_new - coefs.v_offset
            correct_ox_voltages[i] = ox_volts_final

    oxygen = _convert_sbe43_oxygen(
        correct_ox_voltages,
        temperature,
        pressure,
        salinity,
        coefs,
        dvdt_values,
    )
    return oxygen


def _convert_sbe43_oxygen(
    voltage: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
    salinity: np.ndarray,
    coefs: cc.Oxygen43Coefficients,
    dvdt_value: np.ndarray,
):
    """Returns the data after converting it to ml/l.

    voltage is expected to be in volts, temperature in deg c, pressure
    in dbar, and salinity in practical salinity (PSU). All equation
    information comes from the June 2013 revision of the SBE43 manual.
    Expects that hysteresis correction is already performed on the
    incoming voltage, if desired.

    :param voltage: SBE43 voltage
    :param temperature: temperature value converted to deg C
    :param pressure: Converted pressure value from the attached CTD, in
        dbar
    :param salinity: Converted salinity value from the attached CTD, in
        practical salinity PSU
    :param coefs: calibration coefficients for the SBE43 sensor
    :param dvdt_value: derivative value of voltage with respect to time
        at this point. Expected to be 0 if not using Tau correction

    :return: converted Oxygen value, in ml/l
    """

    # Oxygen Solubility equation constants, From SBE43 Manual Appendix A
    a0 = 2.00907
    a1 = 3.22014
    a2 = 4.0501
    a3 = 4.94457
    a4 = -0.256847
    a5 = 3.88767
    b0 = -0.00624523
    b1 = -0.00737614
    b2 = -0.010341
    b3 = -0.00817083
    c0 = -0.000000488682

    ts = np.log((KELVIN_OFFSET_25C - temperature) / (KELVIN_OFFSET_0C + temperature))
    a_term = a0 + a1 * ts + a2 * ts**2 + a3 * ts**3 + a4 * ts**4 + a5 * ts**5
    b_term = salinity * (b0 + b1 * ts + b2 * ts**2 + b3 * ts**3)
    c_term = c0 * salinity**2
    solubility = np.exp(a_term + b_term + c_term)

    # Tau correction
    tau = coefs.tau_20 * np.exp(coefs.d1 * pressure + coefs.d2 * (temperature - 20)) * dvdt_value

    soc_term = coefs.soc * (voltage + coefs.v_offset + tau)
    temp_term = 1.0 + coefs.a * temperature + coefs.b * temperature**2 + coefs.c * temperature**3
    oxygen = (
        soc_term
        * solubility
        * temp_term
        * np.exp((coefs.e * pressure) / (temperature + KELVIN_OFFSET_0C))
    )
    return oxygen


def convert_oxygen_to_mg_per_l(ox_values: np.ndarray):
    """Converts given oxygen values to milligrams/Liter.

    From Application Note 64.

    :param ox_values: oxygen values, already converted to ml/L

    :return: oxygen values converted to milligrams/Liter
    """

    return ox_values * OXYGEN_MLPERL_TO_MGPERL


def convert_oxygen_to_umol_per_kg(ox_values: np.ndarray, potential_density: np.ndarray):
    """Converts given oxygen values to milligrams/kg.

    Note: Sigma-Theta is expected to be calculated via gsw_sigma0,
    meaning is it technically potential density anomaly. Calculating
    using gsw_rho(SA, CT, p_ref = 0) results in actual potential
    density, but this function already does the converison, so values
    will need to have 1000 subtracted from them before being passed into
    this function. The function is done this way to stay matching to
    Application Note 64, but the results of either method are identical.

    :param ox_values: oxygen values, already converted to ml/L
    :param potential_density: potential density (sigma-theta) values.
        Expected to be the same length as ox_values

    :return: oxygen values converted to milligrams/Liter
    """

    oxygen_umolkg = (ox_values * OXYGEN_MLPERL_TO_UMOLPERKG) / (potential_density + 1000)
    return oxygen_umolkg


def convert_eco(
    raw: np.ndarray,
    coefs: cc.ECOCoefficients,
):
    """Converts a raw value for any ECO measurand.

    :param raw: raw counts for digital, raw volts for analog
    :param coefs (ChlorophyllACoefficients): calibration coefficients

    :return: converted ECO measurement in calibration units
    """
    converted = coefs.slope * (raw - coefs.offset)

    return converted


def convert_sbe18_ph(
    raw_ph: np.ndarray,
    temperature: np.ndarray,
    coefs: cc.PH18Coefficients,
):
    """Converts a raw voltage value for pH.

    All equation information comes from application note 18-1

    :param raw_ph: raw output voltage from pH sensor (0-5V)
    :param temperature: temperature value to use for temperature
        compensation in degrees C
    :param coefs: slope and offset for the pH sensor

    :return: converted pH
    """
    ph = 7 + (raw_ph - coefs.offset) / (
        1.98416e-4 * (temperature + KELVIN_OFFSET_0C) * coefs.slope
    )
    return ph


def convert_par_logarithmic(
    volts: np.ndarray,
    coefs: cc.PARCoefficients,
):
    """Converts a raw voltage value for underwater PAR.

    All equation information comes from application note 96

    conversion_factor = 1.0 for units of μmol photons/m2*s

    :param raw_par: raw output voltage from PAR sensor
    :param coefs: calibration coefficients for the PAR sensor

    :return: converted PAR in µmol photons/m2*s
    """
    exponent = (volts - coefs.a0) / coefs.a1
    par = coefs.multiplier * coefs.im * 10**exponent

    return par


def convert_spar_logarithmic(
    volts: np.ndarray,
    coefs: cc.SPARCoefficients,
):
    """Converts a raw voltage value for logarithmic surface PAR.

    All equation information comes from application note 96

    conversion_factor = 1.0 for units of μmol photons/m2*s

    :param volts: raw output voltage from SPAR sensor
    :param coefs: coefficients for the SPAR sensors

    :return: converted surface PAR in µmol photons/m2*s
    """
    exponent = (volts - coefs.a0) / coefs.a1
    spar = coefs.conversion_factor * coefs.im * 10**exponent

    return spar


def convert_spar_linear(
    volts: np.ndarray,
    coefs: cc.SPARCoefficients,
):
    """Converts a raw voltage value for linear surface PAR.

    All equation information comes from application note 96

    conversion_factor = 1.0 for units of μmol photons/m2*s

    :param volts: raw output voltage from SPAR sensor
    :param coefs: coefficients for the SPAR sensors

    :return: converted surface PAR in µmol photons/m2*s
    """
    spar = coefs.im * coefs.a1 * (volts - coefs.a0) * coefs.conversion_factor

    return spar


def convert_spar_biospherical(
    volts: np.ndarray,
    coefs: cc.SPARCoefficients,
):
    """Converts a raw voltage value for biospherical surface PAR.

    All equation information comes from application note 11S

    :param volts: raw output voltage from SPAR sensor
    :param coefs: coefficients for the SPAR sensors

    :return: converted surface PAR in µmol photons/m2*s
    """
    spar = volts * coefs.conversion_factor

    return spar


def convert_nitrate(
    volts: np.ndarray,
    dac_min: float,
    dac_max: float,
    units: Literal["uMNO3", "mgNL"] = "uMNO3",
):
    """Convert SUNA raw voltages to uMNO3 or mgNL

    :param volts: raw output voltage from a SUNA
    :param dac_min: NO3 value that corresponds to v_min
    :param dac_max: NO3 value that corresponds to v_max
    :param units: conversion output units, defaults to 'uMNO3'
    :return: converted nitrate
    """
    v_min = 0.095
    v_max = 4.095
    a1 = (dac_min - dac_max) / (v_min - v_max)
    a0 = dac_max - v_max * a1

    nitrate = a1 * volts + a0

    if units == "mgNL":
        nitrate *= UMNO3_TO_MGNL

    return nitrate


def convert_ph_voltage_counts(ph_counts: np.ndarray):
    """Convert pH voltage counts to a floating point value

    :param ph_counts: pH voltage counts
    :return: pH voltage
    """
    adc_vref = 2.5
    gain = 1
    adc_23bit = 8388608
    ph_volts = adc_vref / gain * (ph_counts / adc_23bit - 1)
    return ph_volts


def _calculate_nernst(temperature: np.ndarray) -> np.ndarray:
    """Calculate the nernst term using natual log

    :param temperature: temperature in kelvin
    :return: the nernst term (J/Coulomb; electrical potential; volts)
    """
    nernst_term = R * temperature * np.log(10) / F
    return nernst_term


def convert_internal_seafet_ph(
    raw_ph: np.ndarray = 0,
    temperature: np.ndarray = 0,
    coefs: cc.PHSeaFETInternalCoefficients = cc.PHSeaFETInternalCoefficients(),
    ph_units: Literal["counts", "volts"] = "counts",
    ph_counts: np.ndarray = None,
):
    """Calculates the internal pH on the total scale given the
    temperature and internal FET voltage

    :param raw_ph: Raw voltage or voltage counts
    :param temperature: Sample temperature
    :param coefs: SeaFET calibration coefficients
    :param ph_units: The units of raw_ph, defaults to 'counts'
    :param ph_counts: Deprecated. pH voltage counts
    :return: calculated pH on the total scale for the SeaFET internal
        reference
    """
    if ph_counts is not None:
        warnings.warn("Deprecated, use raw_ph", DeprecationWarning)
        ph_volts = convert_ph_voltage_counts(ph_counts)
    elif ph_units == "counts":
        ph_volts = convert_ph_voltage_counts(raw_ph)
    else:  # ph_counts == 'volts'
        ph_volts = raw_ph

    nernst = _calculate_nernst(temperature + KELVIN_OFFSET_0C)
    ph = (ph_volts - coefs.kdf0 - coefs.kdf2 * temperature) / nernst
    return ph


def _calculate_thermal_pressure(temperature: np.ndarray, pressure: np.ndarray):
    """ThermPress at in-situ
    Thermpress = (-V_Cl x P + 0.5 K_Cl x P^2)/10F
    Where,
    P: Pressure in 'bar'
    V_Cl: Chloride partial molal volume (in cm^3 mol^-1)
    K_Cl: Chloride partial molal compressibility
    (in cm^3 mol^-1 bar^-1) (can be neglected [4, pg. 876])

    :param temperature: temperature in C
    :param pressure: pressure in bar
    :return: thermal pressure
    """
    # // Partial molal volume and compressibility change for HCl from Millero
    delta_vhcl = _partial_molal_hcl_volume(temperature)

    # // Thermpress term
    thermal_pressure = -delta_vhcl * 0.0242 / (23061 * 1.01) * pressure

    return thermal_pressure


def _total_chloride_in_seawater(salinity: np.ndarray) -> np.ndarray:
    """From SBS application note 99. Calculated as (Dickson et al. 2007)

    :param salinity: Salinity in PSU
    :return: Total chloride in seawater
    """
    # 0.99889  relative concentration of chloride in SW
    # 35.453  [g/mol] molar mass of chloride
    # 1.80655  [ppt, 10^{-3}] Chlorinity to Salinity
    factor_1 = 0.99889 / 35.453
    factor_2 = salinity / 1.80655
    factor_3 = 1 / (1 - 1.005e-3 * salinity)
    total_chloride = factor_1 * factor_2 * factor_3
    return total_chloride


def _total_sulfate_in_seawater(salinity: np.ndarray):
    """From SBS application note 99. Calculated as (Dickson et al. 2007)

    :param salinity: Salinity in PSU
    :return: Total sulfate in seawater
    """
    # 0.14  relative concentration of sulfate in SW
    # 96.062  [g/mol] molar mass of sulfate
    # 1.80655  [ppt, 10^{-3}] Chlorinity to Salinity
    total_sulfate = (0.14 / 96.062) * (salinity / 1.80655)
    return total_sulfate


def _sample_ionic_strength(salinity: np.ndarray) -> np.ndarray:
    """From SBS application note 99. The sample ionic strength is
    calculated as (Dickson et al. 2007)

    :param salinity: Salinity in PSU
    :return: Sample ionic strength
    """
    ionic_strength = (19.924 * salinity) / (1000 - 1.005 * salinity)
    return ionic_strength


def _debye_huckel_constant_for_hcl_activity(temperature: np.ndarray):
    """From SBS application note 99. This constant is calculated as
    (Khoo et al. 1977)

    :param temperature: Temperature in degrees C
    :return: Debye-Huckel constant for activity of HCl
    """
    activity = 3.4286e-6 * temperature**2 + 6.7524e-4 * temperature + 0.49172143
    return activity


def _log_of_hcl_activity_coefficient_of_t(salinity: np.ndarray, temperature: np.ndarray):
    """From SBS application note 99. Calculated as (Khoo et al. 1977)

    :param salinity: Salinity in PSU
    :param temperature: Temperature in degrees C
    :return: Logarithm of HCl activity coefficient as a function of temperature
    """
    i = _sample_ionic_strength(salinity)
    a_dh = _debye_huckel_constant_for_hcl_activity(temperature)
    term_1 = (-a_dh * np.sqrt(i)) / (1 + 1.394 * np.sqrt(i))
    term_2 = (0.08885 - 1.11e-4 * temperature) * i
    log_y_hcl = term_1 + term_2
    return log_y_hcl


def _log_of_hcl_activity_coefficient_of_tp(
    salinity: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
):
    """From SBS application note 99. Calculated as (Johnson et al. 2017)

    :param salinity: Salinity in PSU
    :param temperature: Temperature in degrees C
    :param pressure: Pressure in bar
    :return: Logarithm of HCl activity coefficient as a function of
        temperature and pressure
    """
    log_y_hcl = _log_of_hcl_activity_coefficient_of_t(salinity, temperature)
    v_hcl = _partial_molal_hcl_volume(temperature)
    t_kelvin = temperature + KELVIN_OFFSET_0C
    term_2 = (v_hcl * pressure) / (np.log(10) * R * t_kelvin * 10) / 2
    log_y_hcl_tp = log_y_hcl + term_2
    return log_y_hcl_tp


def _acid_dissociation_constant_of_hso4(salinity: np.ndarray, temperature: np.ndarray):
    """From SBS application note 99. Calculated as (Dickson et al. 2007)

    :param salinity: Salinty in PSU
    :param temperature: Temperature in Kelvin
    :return: _description_
    """
    i = _sample_ionic_strength(salinity)
    term_1 = -4276.1 / temperature + 141.328 - 23.093 * np.log(temperature)
    term_2 = (-13856 / temperature + 324.57 - 47.986 * np.log(temperature)) * np.sqrt(i)
    term_3 = (35474 / temperature - 771.54 + 114.723 * np.log(temperature)) * i
    term_4 = (2698 / temperature) * i**1.5
    term_5 = (1776 / temperature) * i**2
    k_s = (1 - 1.005e-3 * salinity) * np.exp(term_1 + term_2 + term_3 - term_4 + term_5)
    return k_s


def _acid_dissociation_constant_of_hso4_tp(
    salinity: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
):
    """From SBS application note 99. Calculated as (Millero 1982)

    :param salinity: Salinity in PSU
    :param temperature: Temperature in degrees C
    :param pressure: Pressure in bar
    :return: Acid dissociation constant of HSO4
    """
    t_kelvin = temperature + KELVIN_OFFSET_0C
    k_s = _acid_dissociation_constant_of_hso4(salinity, t_kelvin)
    v_bar_s = _partial_molal_hso4_volume(temperature)
    k_bar_s = _hso4_compressibility(temperature)
    exponent = (-v_bar_s * pressure + 0.5 * k_bar_s * pressure**2) / (R * t_kelvin * 10)
    k_stp = k_s * np.exp(exponent)
    return k_stp


def _pressure_response(pressure: np.ndarray, coefs: cc.PHSeaFETExternalCoefficients):
    """The sensor pressure response function from SBS application note
    99.

    :param pressure: Pressure in bar
    :param coefs: External pH coefficients
    :return: The pressure response
    """
    term_1 = coefs.f1 * pressure
    term_2 = coefs.f2 * pressure**2
    term_3 = coefs.f3 * pressure**3
    term_4 = coefs.f4 * pressure**4
    term_5 = coefs.f5 * pressure**5
    term_6 = coefs.f6 * pressure**6
    return term_1 + term_2 + term_3 + term_4 + term_5 + term_6


def _partial_molal_hcl_volume(temperature: np.ndarray):
    """From SBS application note 99. Calculated as (Millero 1983)
    Note: AN99 has a typo and is missing temperature in the second term

    :param temperature: Temperature in degrees C
    :return: Partial Molal Volume of HCl
    """
    volume = 17.85 + 0.1044 * temperature - 1.316e-3 * temperature**2
    return volume


def _partial_molal_hso4_volume(temperature: np.ndarray) -> np.ndarray:
    """From SBS application note 99. Calculated as (Millero 1983)

    :param temperature: Temperature in dgrees C
    :return: Partial Molal Volume of HSO4
    """
    volume = -18.03 + 0.0466 * temperature + 3.16e-4 * temperature**2
    return volume


def _hso4_compressibility(temperature: np.ndarray):
    """From SBS application note 99. Calculated as (Millero 1983)

    :param temperature: Temperature in degrees C
    :return: Compressibility of HSO4
    """
    compressibility = (-4.53 + 0.09 * temperature) / 1000
    return compressibility


def convert_external_seafet_ph(
    raw_ph: np.ndarray = 0,
    temperature: np.ndarray = 0,
    salinity: np.ndarray = 0,
    pressure: np.ndarray = 0,
    coefs: cc.PHSeaFETExternalCoefficients = cc.PHSeaFETExternalCoefficients(),
    ph_units: Literal["counts", "volts"] = "counts",
    formula_version: Literal["legacy", "1.3"] = "1.3",
    ph_counts: np.ndarray = None,
):
    """External pH for the SeaFET, SeapHOx, and Float. From SBS
    Application Note 99 and "Processing BGC-Argo pH data at the DAC
    level"

    https://www.seabird.com/asset-get.download.jsa?id=69833850609
    https://archimer.ifremer.fr/doc/00460/57195/

    :param raw_ph: raw voltage or voltage counts
    :param temperature: Temperature in degrees C
    :param salinity: Salinity in PSU
    :param pressure: Pressure in dbar
    :param coefs: External pH coefficients
    :param ph_units: The units for raw_ph, defaults to "counts"
    :param formula_version: The version of the pH formula, where
        "legacy" refers to the formula used by Fathom v3.0.4 and UCI
        v4.0.x, and "1.3" refers to the version of the Argo pH doc in
        the description
    :return: Total external pH
    """
    if ph_counts is not None:
        warnings.warn("Deprecated, use raw_ph", DeprecationWarning)
        ph_volts = convert_ph_voltage_counts(ph_counts)
    elif ph_units == "counts":
        ph_volts = convert_ph_voltage_counts(raw_ph)
    else:  # ph_counts == 'volts'
        ph_volts = raw_ph

    t_kelvin = temperature + KELVIN_OFFSET_0C
    p_bar = pressure / 10
    f_p = _pressure_response(pressure, coefs)
    nernst = _calculate_nernst(t_kelvin)
    s_t = _total_sulfate_in_seawater(salinity)
    k_stp = _acid_dissociation_constant_of_hso4_tp(salinity, temperature, p_bar)

    term_2 = np.log10(_total_chloride_in_seawater(salinity))
    term_4 = np.log10(1 - 1.005e-3 * salinity)
    term_5 = np.log10(1 + s_t / k_stp)

    if coefs.k2_poly_order == 0:
        k2_poly = Polynomial([coefs.k2])
    else:
        k2_poly = Polynomial([coefs.k2f0, coefs.k2f1, coefs.k2f2, coefs.k2f3])

    eot = k2_poly(pressure) * temperature

    if formula_version == "legacy":
        thermal_pressure = _calculate_thermal_pressure(temperature, p_bar)
        term_1 = (ph_volts - coefs.k0 - eot - f_p - thermal_pressure) / nernst
        term_3 = 2 * _log_of_hcl_activity_coefficient_of_t(salinity, temperature)

    elif formula_version == "1.3":
        term_1 = (ph_volts - coefs.k0 - eot - f_p) / nernst
        term_3 = 2 * _log_of_hcl_activity_coefficient_of_tp(salinity, temperature, p_bar)

    ph = term_1 + term_2 + term_3 - term_4 - term_5
    return ph


def convert_seafet_temperature(raw_temp, coefs: cc.TemperatureSeaFETCoefficients):
    """Converts the raw SeaFET temperature value to ITS-90 Celsius.

    :param raw_temp: raw temperature values
    :return: ITS-90 Celsius.
    """
    temp_log = np.log(raw_temp)

    temp = 1 / (
        ((coefs.tdfa3 * temp_log + coefs.tdfa2) * temp_log + coefs.tdfa1) * temp_log + coefs.tdfa0
    )

    temp_c = temp - KELVIN_OFFSET_0C

    return temp_c


def convert_internal_seafet_temperature(temperature_counts: np.ndarray):
    """Converts the raw internal temperature counts to degrees Celsius

    :param temperature_counts: raw internal temperature counts
    :return: internal temperature in Celsius
    """
    slope = 175.72
    offset = -46.85
    int_16bit = 2**16
    temperature = temperature_counts / int_16bit * slope + offset
    return temperature


def convert_seafet_relative_humidity(humidity_counts: np.ndarray, temperature: np.ndarray):
    """Convert relative humidity counts to percent

    :param humidity_counts: raw relative humidity counts
    :param temperature: converted internal temperature in Celsius
    :return: temperature compensated relative humidity in percent
    """
    slope = 125
    offset = -6
    int_16bit = 2**16
    max_humidity = 119
    temperature_coefficient = -0.15
    temperature_25c = 25

    # Uncompensated relative humidity
    relative_humidity = slope * humidity_counts / int_16bit + offset

    for n, humidity in enumerate(relative_humidity):
        # Theoretically, uncompensated relative humidity can be up to 119%
        if 0 <= humidity < max_humidity:
            relative_humidity[n] = humidity + temperature_coefficient * (
                temperature_25c - temperature[n]
            )

    np.clip(relative_humidity, a_min=0, a_max=100)

    return relative_humidity


def convert_altimeter(
    volts: np.ndarray,
    coefs: cc.AltimeterCoefficients,
):
    """Converts a raw voltage value for altimeter.

    All equation information comes from application note 95

    :param volts: raw output voltage from altimeter sensor
    :param coefs: slope and offset for the altimeter sensors

    :return: converted height in meters
    """
    ALTIMETER_SCALAR = 300

    height = ALTIMETER_SCALAR * volts / coefs.slope - coefs.offset

    return height



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""EOS80 functions to support legacy processing"""

import numpy as np
from scipy import stats


def bouyancy_frequency(
    temp_ITS_subset: np.ndarray,  # TODO: change this to be all lower case for TKIT-75
    salinity_prac_subset: np.ndarray,
    pressure_dbar_subset: np.ndarray,
    gravity: float,
):
    """Calculates an N^2 value (buoyancy frequency) for the given window
    of temperature, salinity, and pressure, at the given latitude.

    Expects temperature as ITS-90 temperature, salinity as practical
    salinity, and pressure as dbar, all of the same length. Performs the
    calculation following the SBE Data Processing formula using E0S-80
    calculations for potential temp and density

    :param temp_ITS_subset: ITS-90 temperature values for the given
        window
    :param salinity_prac_subset: practical salinity values for the given
        window
    :param pressure_dbar_subset: pressure values for the given window
    :param gravity: gravity value

    :return: A single N^2 [Brunt-Väisälä (buoyancy) frequency]
    """

    db_to_pa = 1e4

    # Wrap these as a length-1 array so that GSW accepts them
    pressure_bar = np.array([np.mean(pressure_dbar_subset)])
    temperature_bar = np.array([np.mean(temp_ITS_subset)])
    salinity_bar = np.array([np.mean(salinity_prac_subset)])

    # Compute average density over the window
    # rho_bar0 = gsw.rho(salinity_bar, temperature_bar, pressure_bar)[0]
    rho_bar = density(salinity_bar, temperature_bar, pressure_bar)[0]

    # Use SBE DP (EOS-80) formulas for potential temp and density
    theta = potential_temperature(
        salinity_prac_subset, temp_ITS_subset, pressure_dbar_subset, pressure_bar
    )
    v_vals = 1.0 / density(salinity_prac_subset, theta, pressure_bar)

    # Estimate vertical gradient of specific volume
    dvdp_result = stats.linregress(pressure_dbar_subset, v_vals)

    # Compute EOS-80 N2 combining computed average density and vertical gradient
    # we index into v_bar, alpha_bar, and beta_bar as they are all arrays of len 1
    n2 = 0 - (rho_bar**2 * gravity**2 * dvdp_result.slope / db_to_pa)
    return n2


# TODO: break this function up
def density(s0: np.ndarray, t: np.ndarray, p0: np.ndarray) -> np.ndarray:
    """EOS-80 density calculation.

    This was ported from CSharedCalc::Density()

    :param s0: salinity data
    :param t: temperature data
    :param p0: pressure data

    :return: resulting density data
    """

    b0 = 8.24493e-1
    b1 = -4.0899e-3
    b2 = 7.6438e-5
    b3 = -8.2467e-7
    b4 = 5.3875e-9

    c0 = -5.72466e-3
    c1 = 1.0227e-4
    c2 = -1.6546e-6

    d0 = 4.8314e-4

    a0 = 999.842594
    a1 = 6.793952e-2
    a2 = -9.095290e-3
    a3 = 1.001685e-4
    a4 = -1.120083e-6
    a5 = 6.536332e-9

    fq0 = 54.6746
    fq1 = -0.603459
    fq2 = 1.09987e-2
    fq3 = -6.1670e-5

    g0 = 7.944e-2
    g1 = 1.6483e-2
    g2 = -5.3009e-4

    i0 = 2.2838e-3
    i1 = -1.0981e-5
    i2 = -1.6078e-6

    j0 = 1.91075e-4

    m0 = -9.9348e-7
    m1 = 2.0816e-8
    m2 = 9.1697e-10

    e0 = 19652.21
    e1 = 148.4206
    e2 = -2.327105
    e3 = 1.360477e-2
    e4 = -5.155288e-5

    h0 = 3.239908
    h1 = 1.43713e-3
    h2 = 1.16092e-4
    h3 = -5.77905e-7

    k0 = 8.50935e-5
    k1 = -6.12293e-6
    k2 = 5.2787e-8

    s0, t, p0 = np.broadcast_arrays(s0, t, p0)
    p = p0.copy()
    s = s0.copy()

    t2 = t * t
    t3 = t * t2
    t4 = t * t3
    t5 = t * t4
    s[s <= 0.0] = 0.000001
    s32 = s**1.5
    p /= 10.0
    sigma = (
        a0
        + a1 * t
        + a2 * t2
        + a3 * t3
        + a4 * t4
        + a5 * t5
        + (b0 + b1 * t + b2 * t2 + b3 * t3 + b4 * t4) * s
        + (c0 + c1 * t + c2 * t2) * s32
        + d0 * s * s
    )

    kw = e0 + e1 * t + e2 * t2 + e3 * t3 + e4 * t4
    aw = h0 + h1 * t + h2 * t2 + h3 * t3
    bw = k0 + k1 * t + k2 * t2

    k = (
        kw
        + (fq0 + fq1 * t + fq2 * t2 + fq3 * t3) * s
        + (g0 + g1 * t + g2 * t2) * s32
        + (aw + (i0 + i1 * t + i2 * t2) * s + (j0 * s32)) * p
        + (bw + (m0 + m1 * t + m2 * t2) * s) * p * p
    )

    val = 1 - p / k

    val[val <= 0] = np.nan
    val = sigma / val

    return val


def potential_temperature(
    s: np.ndarray, t0: np.ndarray, p0: np.ndarray, pr: np.ndarray
) -> np.ndarray:
    """EOS-80 potential temperature calculation.

    This was ported from CSharedCalc::PoTemp()

    :param s: sainity data
    :param t0: temperature data
    :param p0: subset pressure data
    :param pr: pressure data

    :return: calculated potential temperature data
    """

    s, t0, p0, pr = np.broadcast_arrays(s, t0, p0, pr)

    p = p0.copy()
    t = t0.copy()
    h = pr - p
    xk = h * adiabatic_temperature_gradient(s, t, p)
    t += 0.5 * xk
    q = xk
    p += 0.5 * h
    xk = h * adiabatic_temperature_gradient(s, t, p)
    t += 0.29289322 * (xk - q)
    q = 0.58578644 * xk + 0.121320344 * q
    xk = h * adiabatic_temperature_gradient(s, t, p)
    t += 1.707106781 * (xk - q)
    q = 3.414213562 * xk - 4.121320344 * q
    p += 0.5 * h
    xk = h * adiabatic_temperature_gradient(s, t, p)
    temp = t + (xk - 2.0 * q) / 6.0

    return temp


def adiabatic_temperature_gradient(
    s: np.ndarray,
    t: np.ndarray,
    p: np.ndarray,
) -> np.ndarray:
    """EOS-80 adiabatic lapse rate calculation.

    This was ported from CSharedCalc::ATG()

    :param s: salinity data
    :param t: temperature data
    :param p: pressure data

    :return: the resulting adiabatic lapse rate
    """

    s, t, p = np.broadcast_arrays(s, t, p)

    ds = s - 35.0
    atg = (
        (
            ((-2.1687e-16 * t + 1.8676e-14) * t - 4.6206e-13) * p
            + (
                (2.7759e-12 * t - 1.1351e-10) * ds
                + ((-5.4481e-14 * t + 8.733e-12) * t - 6.7795e-10) * t
                + 1.8741e-8
            )
        )
        * p
        + (-4.2393e-8 * t + 1.8932e-6) * ds
        + ((6.6228e-10 * t - 6.836e-8) * t + 8.5258e-6) * t
        + 3.5803e-5
    )

    return atg



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""Contains a class and functions to support calculation of temperature
and salinity (TS) contours.
"""

# Classes:
#   ContourData
# Functions:
#   contour_from_t_s_p (np.ndarray, np.ndarray, np.ndarray, float, float, float) -> ContourData
#   contour_from_t_c_p (np.ndarray, np.ndarray, np.ndarray, float, float, float) -> ContourData

# Native imports
from dataclasses import dataclass

# Third-party imports
import gsw
import numpy as np

# Sea-Bird imports

# Internal imports


@dataclass
class ContourData:
    """Container for contour data for a TS plot.

    Example variables are shown next to each field
    """

    x: np.ndarray  # absolute salinity
    y: np.ndarray  # conservative temperature
    z: np.ndarray  # potential density
    x_vec: np.ndarray  # absolute salinity vector
    y_vec: np.ndarray  # conservative temperature vector
    z_mat: np.ndarray  # potential density matrix


def contour_from_t_s_p(
    temperature_C: np.ndarray,  # TODO: change this to be snake_case for TKIT-75
    salinity_PSU: np.ndarray,  # TODO: change this to be snake_case for TKIT-75
    pressure_dbar: np.ndarray,
    min_salinity: float = 0,
    lat: float = 0,
    lon: float = 0,
) -> ContourData:
    """Converts temperature (T), salinity (S), and pressure (P) to
    conservative temperature (CT), absolute salinity (SA), and potential
    density (PD). CT is derived from ITS-90 temperature and practical
    salinity measurements. PD is derived from SA and CT.

    :param temperature_C: Measured temperature in degrees C
    :param salinity_PSU: Measured salinity in practical salinity units
    :param pressure_dbar: Measured pressure in decibars
    :param min_salinity: Minimum salinity to include in contour data.
        Defaults to 0
    :param lat: Used to determine absolute salinity (SA). Defaults to 0
    :param lon: Used to determine absolute salinity (SA). Defaults to 0

    :return: dataclass with xyz data for creating a TS plot
    """

    # Filter out data points where salinity < min_salinity
    salt_mask = salinity_PSU > min_salinity
    temperature_C = temperature_C[salt_mask]
    salinity_PSU = salinity_PSU[salt_mask]
    pressure_dbar = pressure_dbar[salt_mask]

    # Compute TEOS-10 quantities: SA, CT, potential_density
    absolute_salinity = gsw.SA_from_SP(salinity_PSU, pressure_dbar, lon, lat)
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature_C, pressure_dbar)
    potential_density = gsw.rho(
        absolute_salinity, conservative_temperature, 0
    )  # TODO: parameterize reference density

    # Figure out T-S grid boundaries (mins and maxes)
    salt_min = absolute_salinity.min() - (0.01 * absolute_salinity.min())
    salt_max = absolute_salinity.max() + (0.01 * absolute_salinity.max())
    temperature_min = conservative_temperature.min() - (0.1 * conservative_temperature.min())
    temperature_max = conservative_temperature.max() + (0.1 * conservative_temperature.max())

    # Calculate how many grid cells we need in the x and y dimensions
    x_range = round((salt_max - salt_min) / 0.1 + 1, 0)
    y_range = round((temperature_max - temperature_min) + 1, 0)
    x_cells = x_range.astype(int)
    y_cells = y_range.astype(int)

    # Create conservative_temperature and absolute_salinity vectors of appropriate dimensions
    temperature_vector = np.linspace(1, y_range - 1, y_cells) + temperature_min
    salinity_vector = np.linspace(1, x_range - 1, x_cells) * 0.1 + salt_min

    # Loop to fill in density
    potential_density_matrix = np.zeros((y_cells, x_cells))
    for j in range(0, y_cells):
        potential_density_matrix[j, :] = gsw.rho(salinity_vector, temperature_vector[j], 0)

    # Subtract 1000 to convert to sigma-t
    potential_density_matrix -= 1000
    potential_density -= 1000

    contour_data = ContourData(
        x=absolute_salinity,
        y=conservative_temperature,
        z=potential_density,
        x_vec=salinity_vector,
        y_vec=temperature_vector,
        z_mat=potential_density_matrix,
    )

    return contour_data


def contour_from_t_c_p(
    temperature_C: np.ndarray,  # TODO: change this to be snake_case for TKIT-75
    conductivity_mScm: np.ndarray,  # TODO: change this to be snake_case for TKIT-75
    pressure_dbar: np.ndarray,
    min_salinity: float = 0,
    lat: float = 0,
    lon: float = 0,
) -> ContourData:
    """Converts conductivity (C) to salinity (S) then calls
    derive_ct_sa_pd_from_t_s_p to derive conservative temperature (CT),
    absolute salinity (SA), and potential density (PD).

    :param temperature_C: Measured temperature in degrees C
    :param Conductivity_mScm: Measured conductivity in mSiemens/cm
    :param pressure_dbar: Measured pressure in decibar
    :param min_salinity: Minimum salinity to include in contour data.
        Defaults to 0
    :param lat: Used to determine absolute salinity (SA). Defaults to 0
    :param lon: Used to determine absolute salinity (SA). Defaults to 0

    :return: dataclass with xyz data for creating a TS plot
    """
    salinity = gsw.SP_from_C(conductivity_mScm, temperature_C, pressure_dbar)
    return contour_from_t_s_p(temperature_C, salinity, pressure_dbar, min_salinity, lat, lon)
